# fmt: off

from rich import print

from agentica_internal.core.print import *
from pathlib import Path

from agentica_internal.core.debug import enable_rich_tracebacks
from agentica_internal.warpc_transcode.transcript import TranscriptFile



enable_rich_tracebacks()

TRANSCRIPTS_DIR = Path(__file__).parent / "transcripts"


def load_transcript(name: str) -> TranscriptFile:
    path = TRANSCRIPTS_DIR / name
    return TranscriptFile(path)


def generate_combined():

    glob = list(TRANSCRIPTS_DIR.glob("*"))

    for j, path in enumerate(glob):
        print(j, path)
        if not path.is_dir():
            continue
        transcript = TranscriptFile(path)
        transcript.recompute(False)
        if problem := transcript.problems():
            print(problem)
            transcript.pprint(width=80)
            break


def run_combined():
    files = list(TRANSCRIPTS_DIR.glob("*.jsonl"))

    for j, path in enumerate(files):
        # if 'demo_fiddler' not in path.name:
        #     continue
        print(j, path)
        if 'agent' in path.name:
            continue
        if not path.is_file():
            continue
        transcript = TranscriptFile(path)
        transcript.recompute(False)
        problems = transcript.problems()
        if problems:
            print(problems)
            transcript.pprint(width=120)


if __name__ == '__main__':
    generate_combined()
    # run_combined()



# fmt: off

import asyncio
import enum

from agentica_internal.testing import *
from agentica_internal.testing.examples.classes.enums import *
from agentica_internal.warpc.worlds.debug_world import *

def enum_members(cls: enum.EnumType) -> list[tuple[str, str, object]]:
    assert isinstance(cls, enum.EnumType)
    assert issubclass(cls, enum.Enum)
    items = []
    for k, v in cls.__members__.items():
        items.append((k, v.name, v.value))
    items.sort()
    return items

def compare_enums(cls_1: enum.EnumType, cls_2: enum.EnumType):
    members_1 = enum_members(cls_1)
    members_2 = enum_members(cls_2)
    assert members_1 == members_2


async def verify_enum_cls(cls: enum.EnumType):

    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    async with pair:

        A, B = pair.pipes
        cls_b = await B(cls)
        cls_a = await A(cls_b)

        assert cls_a is cls
        assert cls_b is not cls_a

        assert len(cls_a.__members__) == len(cls_b.__members__)
        assert enum_members(cls_a) == enum_members(cls_b)

        for name, member in cls_a.__members__.items():
            member_b = await B(member)
            member_a = await A(member_b)
            assert member_a is member
            assert member_b is not member
            assert member_a.name == member_b.name
            assert member_a.value == member_b.value


async def verify_enum_obj(member: enum.Enum):
    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    async with pair:
        A, B = pair.pipes

        member_b = await B(member)
        member_a = await A(member_b)
        assert member_a is member
        assert member_b is not member
        assert member_a.name == member_b.name
        assert member_a.value == member_b.value


def verify_enum_cls_sync(cls: enum.EnumType):
    asyncio.run(verify_enum_cls(cls))

def verify_enum_obj_sync(obj: enum.Enum):
    asyncio.run(verify_enum_obj(obj))


def verify_enum_classes():
    run_object_tests(verify_enum_cls_sync, ENUM_CLASSES)

def verify_enum_objects():
    run_object_tests(verify_enum_obj_sync, ENUM_INSTANCES)

if __name__ == '__main__':
    verify_enum_classes()
    verify_enum_objects()



# fmt: off

from enum import IntEnum


__all__ = [
    'SysAtomIDs',
    'SysObjectIDs'
]


################################################################################

# this mirrors core.cpython.ids

# in javascript: not objects, but special atomic values
class SysAtomIDs(IntEnum):
    object                  = -1
    string                  = -2
    number                  = -3
    boolean                 = -4
    void                    = -10
    undefined               = -10
    null                    = -10

# ordinary objects
class SysObjectIDs(IntEnum):

    # Classes
    Object                 = -1
    String                 = -2
    Number                 = -3
    Boolean                = -4
    Array                  = -5
    Map                    = -6
    Set                    = -7
    Function               = -8
    Promise                = -9
    None_                  = -10
    Struct                 = -12
    Tuple                  = -13
    Never                  = -14
    Literal                = -15

    # Exceptions
    Error                  = -50
    AttributeError         = -51
    KeyError               = -52
    ValueError             = -53
    TypeError              = -54

    # Objects
    console                = -101
    Math                   = -102
    JSON                   = -103

    # Functions
    parseInt               = -201
    parseFloat             = -202
    isNaN                  = -203
    isFinite               = -204
    magicRepr              = -301
    console_log            = -302
    magicStackTrace        = -303
    magicLen               = -304
    magicKeys              = -305
    magicInstanceof        = -310  # NOT IMPLEMENTED
    Object_entries         = -401
    Object_keys            = -402
    Object_is              = -403
    Reflect_getPrototypeOf = -501
    Reflect_isExtensible   = -502
    Reflect_has            = -503
    Reflect_get            = -504  # NOT IMPLEMENTED


# in 3.13 we could use ._add_alias_ instead
def add_aliases(members: dict):
    aliases = {
        k.strip('_').replace('_', '.'): v
        for k, v in members.items()
        if '_' in k
    }
    members.update(aliases)

add_aliases(SysObjectIDs._member_map_)
del add_aliases



# fmt: off

from .repl_abc import AbstractRepl
from .repl import BaseRepl
from .repl_alias import Vars, Scope, VarKeys, REPL_VAR
from .repl_code import ReplCode, CompileOpts
from .repl_vars import ReplVars
from .repl_vars_delta import VarsDelta
from .repl_eval_data import ReplEvaluationData, ReplError
from .repl_eval_info import ReplEvaluationInfo
from .repl_traceback import register_repl_path
from .repl_callable_info import ReplCallableInfo
from .repl_var_info import VarKind, ReplVarInfo, var_kind
from .repl_session_info import ReplSessionInfo, ReplResourcesInfo, ReplRole, VALID_REPL_ROLES

__all__ = [
    'AbstractRepl',
    'BaseRepl',
    'ReplCode',
    'Vars',
    'VarKeys',
    'Scope',
    'CompileOpts',
    'ReplVars',
    'VarsDelta',
    'ReplEvaluationData',
    'ReplError',
    'ReplEvaluationInfo',
    'ReplCallableInfo',
    'VarKind',
    'ReplVarInfo',
    'ReplSessionInfo',
    'ReplResourcesInfo',
    'ReplRole',
    'VALID_REPL_ROLES',
    'register_repl_path',
    'var_kind',
    'REPL_VAR',
]



# ruff: noqa
# fmt: off

import pytest
from agentica_internal.testing.warpc.virtual_file_tests import *


@pytest.mark.asyncio
async def test_virtual_path():
    await verify_virtual_path()

@pytest.mark.asyncio
async def test_text_write_via_virtual_file():
    await verify_text_write_via_virtual_file()

@pytest.mark.asyncio
async def test_text_read_via_virtual_file():
    await verify_text_read_via_virtual_file()



from typing import Self

from ...core.color import DIM, RED

__all__ = [
    'TestCounts',
]


class TestCounts:
    num_pass = 0
    num_fail = 0
    num_new = 0
    num_err = 0

    def __init__(self, num_pass: int = 0, num_fail: int = 0, num_new: int = 0, num_err: int = 0):
        self.num_pass = num_pass
        self.num_fail = num_fail
        self.num_new = num_new
        self.num_err = num_err

    def reset(self):
        self.num_pass = 0
        self.num_fail = 0
        self.num_new = 0
        self.num_err = 0

    def add_from(self, other: Self):
        self.num_pass += other.num_pass
        self.num_fail += other.num_fail
        self.num_new += other.num_new
        self.num_err += other.num_err

    def nonpassing_count(self) -> int:
        return self.num_fail + self.num_new + self.num_err

    def counts(self) -> tuple[int, int, int, int]:
        return self.num_pass, self.num_fail, self.num_new, self.num_err

    def __str__(self) -> str:
        num_pass, num_fail, num_new, num_err = self.counts()
        summary = f'{num_pass} passed, {num_fail} failed, {num_new} created, {num_err} errors'
        sum_col = RED if num_fail or num_new or num_err else DIM
        return sum_col(summary)



# System resources: stable negative IDs within each frame
ATOMIC_IDS = {
    "object": -1,
    "string": -2,
    "number": -3,
    "boolean": -4,
    "void": -10,
    "undefined": -10,
    "null": -10,
}

BUILTIN_UNI_IDS = {
    # Classes
    "Object": -1,
    "String": -2,
    "Number": -3,
    "Boolean": -4,
    "Array": -5,
    "Map": -6,
    "Set": -7,
    "Function": -8,
    "Future": -9,
    "None": -10,
    "Tuple": -13,
    "Never": -14,
    "Literal": -15,
    "Enum": -16,
    "TYPE": -42,
    "ClassVar": -43,
    # Exceptions
    "Error": -50,
    "AttributeError": -51,
    "KeyError": -52,
    "ValueError": -53,
    "TypeError": -54,
    # Objects
    "console": -101,
    "Math": -102,
    "JSON": -103,
    # Functions
    "parseInt": -201,
    "parseFloat": -202,
    "isNaN": -203,
    "isFinite": -204,
    "magicRepr": -301,
    "console.log": -302,
    "magicStackTrace": -303,
    "magicLen": -304,
    "magicKeys": -305,
    "Object.entries": -401,
    "Object.keys": -402,
    "Object.is": -403,
    "Reflect.getPrototypeOf": -501,
    "Reflect.isExtensible": -502,
    "Reflect.has": -503,
}



from sys import getrecursionlimit, setrecursionlimit

__all__ = ['no_recursion_limit']


################################################################################


class NoRecursionLimit:
    limit: int = 0
    count: int = 0

    def __init__(self):
        self.limit = 0
        self.count = 0

    def __enter__(self):
        self.count += 1
        if self.count == 1:
            self.limit = getrecursionlimit()
            setrecursionlimit(MAX_RECURSION_LIMIT)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.count -= 1
        if self.count == 0:
            setrecursionlimit(self.limit)


no_recursion_limit = NoRecursionLimit()


################################################################################

MAX_RECURSION_LIMIT = getrecursionlimit()



from functools import wraps

from agentica_internal.core.mixin import mixin

from . import a
from .a import *

__all__ = 'foo', 'bar', 'x', 'decorated_caller', 'helper', 'get_registry', 'is_feature_b'

mixin(a)

x = 2


def foo() -> str:
    return 'b.foo'


def helper() -> str:
    return 'b.helper'


def is_feature_b() -> bool:
    return True


def closure_decorator(func):
    """wraps a generator, capturing it in a closure"""

    @wraps(func)
    def wrapper(*args, **kwargs):
        gen = func(*args, **kwargs)
        results = list(gen)
        return results

    return wrapper


@closure_decorator
def decorated_caller():
    """generator that calls helper() and foo()"""
    yield helper()
    yield foo()



# fmt: off

import enum
import sys

__all__ = [
    'SysModule',
    'SysClassID',
    'SysObjectID',
    'SysFunctionID',
    'SYS_MOD_TABLE',
    'SYS_CLASS_ID_DICT',
    'SYS_OBJECT_ID_DICT',
    'SYS_FUNCTION_ID_DICT',
]

###############################################################################

PY13 = sys.version_info >= (3, 13)


class SysModule(enum.StrEnum):
    B = 'builtins'
    S = 'types'
    X = 'agentica_internal.cpython.iters'
    Z = 'agentica_internal.core.sentinels'
    J = 'agentica_internal.javascript'
    JS = 'json'
    A = 'abc'
    T = 'typing'
    C = 'collections'
    CA = 'collections.abc'
    CF = 'concurrent.futures'
    AI = 'asyncio'
    AIE = 'asyncio.exceptions'
    AIQ = 'asyncio.queues'
    EN = 'enum'
    DC = 'dataclasses'
    IT = 'itertools'
    FT = 'functools'
    IN = 'inspect'
    IO = 'io'
    RE = 're'
    OS = 'os'
    OP = 'operator'
    DT = 'datetime'
    RAND = 'random'


SYS_MOD_TABLE: list[tuple[int, str, str]]
SYS_MOD_TABLE = [(i, k, str(v)) for i, (k, v) in enumerate(SysModule.__members__.items())]


###############################################################################


class SysClassID(enum.IntEnum):
    """
    Names some builtin types and orders them in a sane way.
    Various types are grouped in ranges that are meaningful in hex.

    These enums can be processed to build tables.

    The keys here are prefixed by the module in which the type can be found; see ModuleID.

    The hex ranges used are:

    00X | immutable datums  | int, bool, ...
    01X | containers        | list, tuple, ...
    02X | collections.*     | defaultdict, deque, ...
    03X | proxy containers  | MappingProxyType, ...
    04X | iterators         | types of iter([]), iter({}), ...
    05X | callables         | FunctionType, MethodType,...
    06x | types.*           | ModuleType, FrameType, ...
    07x | builtins.*        | property, staticmethod, map, filter, ...

    0EX | typing            | Any, NewType, ...

    1XX | abcs              | ABC, Iterable, ...
    2XX | protocols         | Protocol, SupportsInt, ...
    30X | enums             | Enum, EnumInt, ...
    31X | dataclasses       | Field, InitVar, ...
    32X | re                | Pattern, Match
    33X | datetime          | date, datetime, time, ...
    40X | itertools         | chain, ...
    50X | functools         | partial, partialmethod, ...
    51X | operator          | partial, partialmethod, ...
    6XX | io                | IOBase, FileIO, ...

    EXX | typing internals  | Any, NewType, ...
    FXX | exceptions        | BaseException, Exception, ...
    """

    B_bool                           = 0x01   # immutable datums
    B_int                            = 0x02
    B_float                          = 0x03
    B_complex                        = 0x04
    B_str                            = 0x05
    B_bytes                          = 0x06
    S_NoneType                       = 0x0A
    S_EllipsisType                   = 0x0B
    S_NotImplementedType             = 0x0C
    B_tuple                          = 0x10
    B_list                           = 0x11   # containers
    B_set                            = 0x12
    B_dict                           = 0x13
    B_frozenset                      = 0x14
    B_bytearray                      = 0x15
    B_memoryview                     = 0x16
    C_ChainMap                       = 0x20   # collections
    C_Counter                        = 0x21
    C_defaultdict                    = 0x22
    C_deque                          = 0x23
    C_OrderedDict                    = 0x24
    S_MappingProxyType               = 0x30   # proxy containers
    if PY13:
        X_FrameLocalsProxyType       = 0x31
    X_DictKeysType                   = 0x32
    X_DictValuesType                 = 0x33
    X_DictItemsType                  = 0x34
    X_DictKeysIterType               = 0x40   # iterators
    X_DictValuesIterType             = 0x41
    X_DictItemsIterType              = 0x42
    X_ListIterType                   = 0x43
    X_TupleIterType                  = 0x44
    X_SetIterType                    = 0x45
    X_ReversedListIterType           = 0x47
    X_StrIterType                    = 0x4A
    X_BytesIterType                  = 0x4B
    X_ByteArrayIterType              = 0x4C
    X_CallableIterType               = 0x4D
    S_FunctionType                   = 0x50   # callable CPython types
    S_BuiltinFunctionType            = 0x51
    S_MethodType                     = 0x52
    S_MethodDescriptorType           = 0x53
    S_WrapperDescriptorType          = 0x54
    S_MethodWrapperType              = 0x55
    S_ClassMethodDescriptorType      = 0x56
    S_CodeType                       = 0x57
    S_ModuleType                     = 0x60   # core CPython types
    S_SimpleNamespace                = 0x61
    S_CellType                       = 0x64
    S_CoroutineType                  = 0x65
    S_GeneratorType                  = 0x66
    S_AsyncGeneratorType             = 0x67
    S_GetSetDescriptorType           = 0x68
    S_MemberDescriptorType           = 0x69
    S_FrameType                      = 0x6A
    S_TracebackType                  = 0x6B
    B_property                       = 0x70   # other builtin types
    B_staticmethod                   = 0x71
    B_classmethod                    = 0x72
    B_super                          = 0x73
    B_slice                          = 0x74
    B_enumerate                      = 0x75
    B_range                          = 0x76
    B_filter                         = 0x77
    B_map                            = 0x78
    B_zip                            = 0x79
    B_reversed                       = 0x80
    T_Any                            = 0xE0   # public typing.* types
    T_Generic                        = 0xE1
    # S_UnionType                    = 0xE2
    # S_GenericAlias                 = 0xE3
    T_ForwardRef                     = 0xE4
    T_NewType                        = 0xE5
    T_TypeVar                        = 0xE6
    T_TextIO                         = 0xE7
    T_BinaryIO                       = 0xE8
    B_object                         = 0xF0   # basic
    B_type                           = 0xFF
    A_ABC                            = 0x100  # ABCs
    CA_Hashable                      = 0x101
    CA_Awaitable                     = 0x102
    CA_Coroutine                     = 0x103
    CA_AsyncIterable                 = 0x104
    CA_AsyncIterator                 = 0x105
    CA_AsyncGenerator                = 0x106
    CA_Iterable                      = 0x107
    CA_Iterator                      = 0x108
    CA_Reversible                    = 0x109
    CA_Generator                     = 0x10A
    CA_Sized                         = 0x10B
    CA_Container                     = 0x10C
    CA_Collection                    = 0x10D
    CA_Buffer                        = 0x10E
    CA__CallableGenericAlias         = 0x10F
    CA_Callable                      = 0x110
    CA_Set                           = 0x111
    CA_MutableSet                    = 0x112
    CA_Mapping                       = 0x114
    CA_MappingView                   = 0x115
    CA_KeysView                      = 0x116
    CA_ItemsView                     = 0x117
    CA_ValuesView                    = 0x118
    CA_MutableMapping                = 0x119
    CA_Sequence                      = 0x11A
    CA_ByteString                    = 0x11B
    CA_MutableSequence               = 0x11C
    A_ABCMeta                        = 0x1F0
    T_Protocol                       = 0x200  # protocols
    T_SupportsInt                    = 0x201
    T_SupportsFloat                  = 0x202
    T_SupportsComplex                = 0x203
    T_SupportsBytes                  = 0x204
    T_SupportsIndex                  = 0x205
    T_SupportsAbs                    = 0x206
    T_SupportsRound                  = 0x207
    EN_Enum                          = 0x300  # enums
    EN_IntEnum                       = 0x301
    EN_StrEnum                       = 0x302
    EN_Flag                          = 0x303
    EN_ReprEnum                      = 0x304
    if PY13:
        EN_EnumDict                  = 0x30C  # internal types
    EN_EnumType                      = 0x30D
    EN_auto                          = 0x30E
    # EN_EnumMeta                      = 0x30F
    DC_Field                         = 0x310  # dataclasses
    DC_InitVar                       = 0x311
    DC__DataclassParams              = 0x312
    RE_Pattern                       = 0x320  # re
    RE_Match                         = 0x321
    DT_date                          = 0x330  # datetime
    DT_datetime                      = 0x331
    DT_time                          = 0x332
    DT_timedelta                     = 0x333
    DT_tzinfo                        = 0x334
    DT_timezone                      = 0x335
    IT_accumulate                    = 0x400  # itertools
    IT_batched                       = 0x401
    IT_chain                         = 0x402
    IT_combinations                  = 0x403
    IT_combinations_with_replacement = 0x404
    IT_compress                      = 0x405
    IT_count                         = 0x406
    IT_cycle                         = 0x407
    IT_dropwhile                     = 0x408
    IT_filterfalse                   = 0x409
    IT_groupby                       = 0x40A
    IT_islice                        = 0x40B
    IT_pairwise                      = 0x40C
    IT_permutations                  = 0x40D
    IT_product                       = 0x40E
    IT_repeat                        = 0x40F
    IT_starmap                       = 0x410
    IT_takewhile                     = 0x411
    IT_zip_longest                   = 0x412
    FT_partial                       = 0x500  # functools
    FT_partialmethod                 = 0x501
    FT_singledispatchmethod          = 0x502
    FT_cached_property               = 0x503
    OP_attrgetter                    = 0x510  # operator types
    OP_itemgetter                    = 0x511
    OP_methodcaller                  = 0x512
    IO_IOBase                        = 0x600  # io
    IO_RawIOBase                     = 0x601
    IO_FileIO                        = 0x602
    IO_BytesIO                       = 0x603
    IO_StringIO                      = 0x604
    IO_BufferedIOBase                = 0x605
    IO_BufferedReader                = 0x606
    IO_BufferedWriter                = 0x607
    IO_BufferedRWPair                = 0x608
    IO_BufferedRandom                = 0x609
    IO_TextIOBase                    = 0x610
    IO_TextIOWrapper                 = 0x611
    IN_BoundArguments                = 0x700  # inspect
    IN_BufferFlags                   = 0x701
    IN_Parameter                     = 0x702
    IN_Signature                     = 0x703
    IN__void                         = 0x7F0
    IN__empty                        = 0x7F1
    IN__ParameterKind                = 0x7F2
    AI_Future                        = 0x800  # asyncio
    AI_Task                          = 0x801
    AI_Event                         = 0x802
    AIQ_Queue                        = 0x803
    AI_AbstractEventLoop             = 0x810
    OS_stat_result                   = 0x900  # os
    RAND_Random                      = 0xA00  # random
    RAND_SystemRandom                = 0xA01
    T__SpecialForm                   = 0xE11  # internal typing.* types / metas
    T__BaseGenericAlias              = 0xE12
    T__GenericAlias                  = 0xE13
    T__SpecialGenericAlias           = 0xE14
    T__DeprecatedGenericAlias        = 0xE15
    T__CallableGenericAlias          = 0xE16
    T__CallableType                  = 0xE17
    T__TupleType                     = 0xE18
    T__UnionGenericAlias             = 0xE19
    T__LiteralGenericAlias           = 0xE1A
    T__ConcatenateGenericAlias       = 0xE1B
    T__UnpackGenericAlias            = 0xE1C
    T__TypingEllipsis                = 0xE1D
    T__AnnotatedAlias                = 0xE1F
    T__IdentityCallable              = 0xE20
    T__Final                         = 0xE22
    T__NotIterable                   = 0xE23
    T_NamedTupleMeta                 = 0xEF0
    T__AnyMeta                       = 0xEF1
    T__TypedDictMeta                 = 0xEF2
    T__ProtocolMeta                  = 0xEF3
    B_BaseException                  = 0xF00  # builtin exceptions
    B_Exception                      = 0xF01
    B_GeneratorExit                  = 0xF02
    B_KeyboardInterrupt              = 0xF03
    B_SystemExit                     = 0xF04
    B_StopIteration                  = 0xF05
    B_OSError                        = 0xF06
    B_ArithmeticError                = 0xF07
    B_AssertionError                 = 0xF08
    B_AttributeError                 = 0xF09
    B_BufferError                    = 0xF0A
    B_EOFError                       = 0xF0B
    B_ImportError                    = 0xF0C
    B_LookupError                    = 0xF0D
    B_MemoryError                    = 0xF0E
    B_NameError                      = 0xF0F
    B_ReferenceError                 = 0xF10
    B_RuntimeError                   = 0xF11
    B_StopAsyncIteration             = 0xF12
    B_SyntaxError                    = 0xF13
    B_SystemError                    = 0xF14
    B_TypeError                      = 0xF15
    B_ValueError                     = 0xF16
    B_FloatingPointError             = 0xF17
    B_OverflowError                  = 0xF18
    B_ZeroDivisionError              = 0xF19
    B_ModuleNotFoundError            = 0xF1A
    B_IndexError                     = 0xF1B
    B_KeyError                       = 0xF1C
    B_UnboundLocalError              = 0xF1D
    B_BlockingIOError                = 0xF1E
    B_ChildProcessError              = 0xF1F
    B_ConnectionError                = 0xF20
    B_BrokenPipeError                = 0xF21
    B_ConnectionAbortedError         = 0xF22
    B_ConnectionRefusedError         = 0xF23
    B_ConnectionResetError           = 0xF24
    B_FileExistsError                = 0xF25
    B_FileNotFoundError              = 0xF26
    B_InterruptedError               = 0xF27
    B_IsADirectoryError              = 0xF28
    B_NotADirectoryError             = 0xF29
    B_PermissionError                = 0xF2A
    B_ProcessLookupError             = 0xF2B
    B_TimeoutError                   = 0xF2C
    B_NotImplementedError            = 0xF2D
    B_RecursionError                 = 0xF2E
    B_IndentationError               = 0xF2F
    B_TabError                       = 0xF30
    B_UnicodeError                   = 0xF31
    B_UnicodeDecodeError             = 0xF32
    B_UnicodeEncodeError             = 0xF33
    B_UnicodeTranslateError          = 0xF34
    B_Warning                        = 0xF35
    B_UserWarning                    = 0xF36
    B_DeprecationWarning             = 0xF37
    B_SyntaxWarning                  = 0xF38
    B_RuntimeWarning                 = 0xF39
    B_FutureWarning                  = 0xF3A
    B_PendingDeprecationWarning      = 0xF3B
    B_ImportWarning                  = 0xF3C
    B_UnicodeWarning                 = 0xF3D
    B_BytesWarning                   = 0xF3E
    B_ResourceWarning                = 0xF3F
    AI_CancelledError                = 0xFA1
    AI_InvalidStateError             = 0xFA2
    CF_CancelledError                = 0xFA4
    CF_BrokenExecutor                = 0xFA5  # wrong module?
    JS_JSONDecodeError               = 0xFA6
    IO_UnsupportedOperation          = 0xFA0
    DC_FrozenInstanceError           = 0xFA1
    IN_ClassFoundException           = 0xFA2
    IN_EndOfBlock                    = 0xFA3


SYS_CLASS_ID_DICT: dict[str, int] = {v._name_: v._value_ for v in SysClassID}

###############################################################################


class SysObjectID(enum.IntEnum):
    B_None                           = 0x00
    B_NotImplemented                 = 0x01
    B_Ellipsis                       = 0x02  # `typing.X` equivalents of `builtins.X`
    T_Tuple                          = 0x10
    T_List                           = 0x11
    T_Set                            = 0x12
    T_Dict                           = 0x13
    T_FrozenSet                      = 0x14  # `typing.X` equivalents of `builtins.X`
    T_ChainMap                       = 0x20
    T_Counter                        = 0x21
    T_DefaultDict                    = 0x22
    T_Deque                          = 0x23
    T_OrderedDict                    = 0x24
    T_Type                           = 0xFF
    T_Hashable                       = 0x101  # `typing.X` equivalents of `collections.abc.X`
    T_Awaitable                      = 0x102
    T_Coroutine                      = 0x103
    T_AsyncIterable                  = 0x104
    T_AsyncIterator                  = 0x105
    T_AsyncGenerator                 = 0x106
    T_Iterable                       = 0x107
    T_Iterator                       = 0x108
    T_Reversible                     = 0x109
    T_Generator                      = 0x10A
    T_Sized                          = 0x10B
    T_Container                      = 0x10C
    T_Collection                     = 0x10D
    T_Callable                       = 0x110
    T_MutableSet                     = 0x112
    T_Mapping                        = 0x114
    T_MappingView                    = 0x115
    T_KeysView                       = 0x116
    T_ItemsView                      = 0x117
    T_ValuesView                     = 0x118
    T_MutableMapping                 = 0x119
    T_Sequence                       = 0x11A
    T_ByteString                     = 0x11B
    T_MutableSequence                = 0x11C

    RAND__inst                       = 0x200  # random default state

    DC__HAS_DEFAULT_FACTORY          = 0xE00  # dataclass sentinels
    DC_MISSING                       = 0xE01
    DC_KW_ONLY                       = 0xE02
    DC__EMPTY_METADATA               = 0xE03
    DC__FIELD                        = 0xE04
    DC__FIELD_CLASSVAR               = 0xE05
    DC__FIELD_INITVAR                = 0xE06
    IN_POSITIONAL_ONLY               = 0xE10  # inspect sentinels
    IN_POSITIONAL_OR_KEYWORD         = 0xE11
    IN_VAR_POSITIONAL                = 0xE12
    IN_KEYWORD_ONLY                  = 0xE13
    IN_VAR_KEYWORD                   = 0xE14
    Z_PENDING                        = 0xEE0
    Z_CLOSED                         = 0xEE1
    Z_CANCELED                       = 0xEE2
    Z_ERRORED                        = 0xEE3
    Z_ARG_DEFAULT                    = 0xEE4
    IN__is_coroutine_mark            = 0xC00
    IN__is_coroutine_marker          = 0xC00  # I'm SURE this existed in 3.12.0
    T_NoReturn                       = 0xF00  # special forms
    T_Never                          = 0xF01
    T_Self                           = 0xF02
    T_LiteralString                  = 0xF03
    T_Union                          = 0xF04
    T_Optional                       = 0xF05
    T_Unpack                         = 0xF06
    T_Literal                        = 0xF07
    T_TypeAlias                      = 0xF08
    T_Concatenate                    = 0xF09
    T_ClassVar                       = 0xF10
    T_Final                          = 0xF11
    T_Required                       = 0xF12
    T_NotRequired                    = 0xF13
    if PY13:
        T_ReadOnly                   = 0xF14
    T_TypeGuard                      = 0xF15
    if PY13:
        T_TypeIs                     = 0xF16
    J_Number                         = 0xFF0  # javascript equivalents


SYS_OBJECT_ID_DICT: dict[str, int] = {v._name_: v._value_ for v in SysObjectID}

###############################################################################


class SysFunctionID(enum.IntEnum):
    B_abs                            = 0x0000
    B_aiter                          = 0x0001
    B_all                            = 0x0002
    B_anext                          = 0x0003
    B_any                            = 0x0004
    B_ascii                          = 0x0005
    B_bin                            = 0x0006
    B_breakpoint                     = 0x0007
    B_callable                       = 0x0008
    B_chr                            = 0x0009
    B_compile                        = 0x000A
    B_delattr                        = 0x000B
    B_dir                            = 0x000C
    B_divmod                         = 0x000D
    B_eval                           = 0x000E
    B_exec                           = 0x000F
    B_format                         = 0x0010
    B_getattr                        = 0x0011
    B_globals                        = 0x0012
    B_hasattr                        = 0x0013
    B_hash                           = 0x0014
    B_hex                            = 0x0015
    B_id                             = 0x0016
    B_input                          = 0x0017
    B_isinstance                     = 0x0018
    B_issubclass                     = 0x0019
    B_iter                           = 0x001A
    B_len                            = 0x001B
    B_locals                         = 0x001C
    B_max                            = 0x001D
    B_min                            = 0x001E
    B_next                           = 0x001F
    B_oct                            = 0x0020
    B_open                           = 0x0021
    B_ord                            = 0x0022
    B_pow                            = 0x0023
    B_print                          = 0x0024
    B_repr                           = 0x0025
    B_round                          = 0x0026
    B_setattr                        = 0x0027
    B_sorted                         = 0x0028
    B_sum                            = 0x0029
    B_vars                           = 0x002A
    S_coroutine                      = 0x0100
    S_get_original_bases             = 0x0101
    S_new_class                      = 0x0102
    S_prepare_class                  = 0x0103
    S_resolve_bases                  = 0x0104
    A_abstractmethod                 = 0x0300
    A_get_cache_token                = 0x0301
    A_update_abstractmethods         = 0x0302
    T_NamedTuple                     = 0x0400
    T_TypedDict                      = 0x0401
    T_assert_never                   = 0x0402
    T_assert_type                    = 0x0403
    T_cast                           = 0x0404
    T_clear_overloads                = 0x0405
    T_dataclass_transform            = 0x0406
    T_final                          = 0x0407
    T_get_args                       = 0x0408
    T_get_origin                     = 0x0409
    T_get_overloads                  = 0x040A
    T_get_type_hints                 = 0x040B
    T_is_typeddict                   = 0x040C
    T_no_type_check                  = 0x040D
    T_no_type_check_decorator        = 0x040E
    T_overload                       = 0x040F
    T_override                       = 0x0410
    T_reveal_type                    = 0x0411
    T_runtime_checkable              = 0x0412
    C_namedtuple                     = 0x0500
    AI_all_tasks                     = 0x0700
    AI_as_completed                  = 0x0701
    AI_create_eager_task_factory     = 0x0702
    AI_create_subprocess_exec        = 0x0703
    AI_create_subprocess_shell       = 0x0704
    AI_create_task                   = 0x0705
    AI_current_task                  = 0x0706
    AI_eager_task_factory            = 0x0707
    AI_ensure_future                 = 0x0708
    AI_gather                        = 0x0709
    AI_get_child_watcher             = 0x070A
    AI_get_event_loop                = 0x070B
    AI_get_event_loop_policy         = 0x070C
    AI_get_running_loop              = 0x070D
    AI_iscoroutine                   = 0x070E
    AI_iscoroutinefunction           = 0x070F
    AI_isfuture                      = 0x0710
    AI_new_event_loop                = 0x0711
    AI_open_connection               = 0x0712
    AI_open_unix_connection          = 0x0713
    AI_run                           = 0x0714
    AI_run_coroutine_threadsafe      = 0x0715
    AI_set_child_watcher             = 0x0716
    AI_set_event_loop                = 0x0717
    AI_set_event_loop_policy         = 0x0718
    AI_shield                        = 0x0719
    AI_sleep                         = 0x071A
    AI_start_server                  = 0x071B
    AI_start_unix_server             = 0x071C
    AI_timeout                       = 0x071D
    AI_timeout_at                    = 0x071E
    AI_to_thread                     = 0x071F
    AI_wait                          = 0x0720
    AI_wait_for                      = 0x0721
    AI_wrap_future                   = 0x0722
    EN_global_enum                   = 0x0800
    EN_global_enum_repr              = 0x0801
    EN_global_flag_repr              = 0x0802
    EN_global_str                    = 0x0803
    EN_pickle_by_enum_name           = 0x0804
    EN_pickle_by_global_name         = 0x0805
    EN_unique                        = 0x0806
    DC_asdict                        = 0x0900
    DC_astuple                       = 0x0901
    DC_dataclass                     = 0x0902
    DC_field                         = 0x0903
    DC_fields                        = 0x0904
    DC_is_dataclass                  = 0x0905
    DC_make_dataclass                = 0x0906
    DC_replace                       = 0x0907
    IT_tee                           = 0x0A00
    FT_cache                         = 0x0B00
    FT_cmp_to_key                    = 0x0B01
    FT_lru_cache                     = 0x0B02
    FT_reduce                        = 0x0B03
    FT_singledispatch                = 0x0B04
    FT_total_ordering                = 0x0B05
    FT_update_wrapper                = 0x0B06
    FT_wraps                         = 0x0B07
    IN_classify_class_attrs          = 0x0C00
    IN_cleandoc                      = 0x0C01
    IN_currentframe                  = 0x0C02
    IN_findsource                    = 0x0C03
    IN_formatannotation              = 0x0C04
    IN_formatannotationrelativeto    = 0x0C05
    IN_formatargvalues               = 0x0C06
    IN_get_annotations               = 0x0C07
    IN_getabsfile                    = 0x0C08
    IN_getargs                       = 0x0C09
    IN_getargvalues                  = 0x0C0A
    IN_getasyncgenlocals             = 0x0C0B
    IN_getasyncgenstate              = 0x0C0C
    IN_getattr_static                = 0x0C0D
    IN_getblock                      = 0x0C0E
    IN_getcallargs                   = 0x0C0F
    IN_getclasstree                  = 0x0C10
    IN_getclosurevars                = 0x0C11
    IN_getcomments                   = 0x0C12
    IN_getcoroutinelocals            = 0x0C13
    IN_getcoroutinestate             = 0x0C14
    IN_getdoc                        = 0x0C15
    IN_getfile                       = 0x0C16
    IN_getframeinfo                  = 0x0C17
    IN_getfullargspec                = 0x0C18
    IN_getgeneratorlocals            = 0x0C19
    IN_getgeneratorstate             = 0x0C1A
    IN_getinnerframes                = 0x0C1B
    IN_getlineno                     = 0x0C1C
    IN_getmembers                    = 0x0C1D
    IN_getmembers_static             = 0x0C1E
    IN_getmodule                     = 0x0C1F
    IN_getmodulename                 = 0x0C20
    IN_getmro                        = 0x0C21
    IN_getouterframes                = 0x0C22
    IN_getsource                     = 0x0C23
    IN_getsourcefile                 = 0x0C24
    IN_getsourcelines                = 0x0C25
    IN_indentsize                    = 0x0C26
    IN_isabstract                    = 0x0C27
    IN_isasyncgen                    = 0x0C28
    IN_isasyncgenfunction            = 0x0C29
    IN_isawaitable                   = 0x0C2A
    IN_isbuiltin                     = 0x0C2B
    IN_isclass                       = 0x0C2C
    IN_iscode                        = 0x0C2D
    IN_iscoroutine                   = 0x0C2E
    IN_iscoroutinefunction           = 0x0C2F
    IN_isdatadescriptor              = 0x0C30
    IN_isframe                       = 0x0C31
    IN_isfunction                    = 0x0C32
    IN_isgenerator                   = 0x0C33
    IN_isgeneratorfunction           = 0x0C34
    IN_isgetsetdescriptor            = 0x0C35
    IN_ismemberdescriptor            = 0x0C36
    IN_ismethod                      = 0x0C37
    IN_ismethoddescriptor            = 0x0C38
    IN_ismethodwrapper               = 0x0C39
    IN_ismodule                      = 0x0C3A
    IN_isroutine                     = 0x0C3B
    IN_istraceback                   = 0x0C3C
    IN_markcoroutinefunction         = 0x0C3D
    IN_signature                     = 0x0C3E
    IN_stack                         = 0x0C3F
    IN_trace                         = 0x0C40
    IN_unwrap                        = 0x0C41
    IN_walktree                      = 0x0C42
    IO_open_code                     = 0x0D00
    RE_compile                       = 0x0E00
    RE_escape                        = 0x0E01
    RE_findall                       = 0x0E02
    RE_finditer                      = 0x0E03
    RE_fullmatch                     = 0x0E04
    RE_match                         = 0x0E05
    RE_purge                         = 0x0E06
    RE_search                        = 0x0E07
    RE_split                         = 0x0E08
    RE_sub                           = 0x0E09
    RE_subn                          = 0x0E0A
    if not PY13: # RE_template was removed in Python 3.13
        RE_template                  = 0x0E0B
    OP_abs                           = 0x0F00
    OP_add                           = 0x0F01
    OP_and_                          = 0x0F02
    OP_call                          = 0x0F03
    OP_concat                        = 0x0F04
    OP_contains                      = 0x0F05
    OP_countOf                       = 0x0F06
    OP_delitem                       = 0x0F07
    OP_eq                            = 0x0F08
    OP_floordiv                      = 0x0F09
    OP_ge                            = 0x0F0A
    OP_getitem                       = 0x0F0B
    OP_gt                            = 0x0F0C
    OP_iadd                          = 0x0F0D
    OP_iand                          = 0x0F0E
    OP_iconcat                       = 0x0F0F
    OP_ifloordiv                     = 0x0F10
    OP_ilshift                       = 0x0F11
    OP_imatmul                       = 0x0F12
    OP_imod                          = 0x0F13
    OP_imul                          = 0x0F14
    OP_index                         = 0x0F15
    OP_indexOf                       = 0x0F16
    OP_inv                           = 0x0F17
    OP_invert                        = 0x0F18
    OP_ior                           = 0x0F19
    OP_ipow                          = 0x0F1A
    OP_irshift                       = 0x0F1B
    OP_is_                           = 0x0F1C
    OP_is_not                        = 0x0F1D
    OP_isub                          = 0x0F1E
    OP_itruediv                      = 0x0F1F
    OP_ixor                          = 0x0F20
    OP_le                            = 0x0F21
    OP_length_hint                   = 0x0F22
    OP_lshift                        = 0x0F23
    OP_lt                            = 0x0F24
    OP_matmul                        = 0x0F25
    OP_mod                           = 0x0F26
    OP_mul                           = 0x0F27
    OP_ne                            = 0x0F28
    OP_neg                           = 0x0F29
    OP_not_                          = 0x0F2A
    OP_or_                           = 0x0F2B
    OP_pos                           = 0x0F2C
    OP_pow                           = 0x0F2D
    OP_rshift                        = 0x0F2E
    OP_setitem                       = 0x0F2F
    OP_sub                           = 0x0F30
    OP_truediv                       = 0x0F31
    OP_truth                         = 0x0F32
    OP_xor                           = 0x0F33
    RAND_seed                        = 0x1000  # random bound methods
    RAND_random                      = 0x1001
    RAND_uniform                     = 0x1002
    RAND_triangular                  = 0x1003
    RAND_randint                     = 0x1004
    RAND_choice                      = 0x1005
    RAND_randrange                   = 0x1006
    RAND_sample                      = 0x1007
    RAND_shuffle                     = 0x1008
    RAND_choices                     = 0x1009
    RAND_normalvariate               = 0x100A
    RAND_lognormvariate              = 0x100B
    RAND_expovariate                 = 0x100C
    RAND_vonmisesvariate             = 0x100D
    RAND_gammavariate                = 0x100E
    RAND_gauss                       = 0x100F
    RAND_betavariate                 = 0x1010
    RAND_binomialvariate             = 0x1011
    RAND_paretovariate               = 0x1012
    RAND_weibullvariate              = 0x1013
    RAND_getstate                    = 0x1014
    RAND_setstate                    = 0x1015
    RAND_getrandbits                 = 0x1016
    RAND_randbytes                   = 0x1017

SYS_FUNCTION_ID_DICT: dict[str, int] = {v._name_: v._value_ for v in SysFunctionID}


# the above is generated by:
def _generate_function_ids():
    from importlib import import_module

    from agentica_internal.cpython.alias import CALLABLES as FUNC_TYPES

    seen = set()
    see = seen.add
    for i, prefix, name in SYS_MOD_TABLE:
        mod = import_module(name)
        j = i << 8
        items = list(vars(mod).items())
        exports = getattr(mod, '__all__', None)
        items.sort()
        for k, v in items:
            if k.startswith('_'):
                continue
            if exports and k not in exports:
                continue
            if isinstance(v, FUNC_TYPES):
                i = id(v)
                if i in seen:
                    continue
                see(i)
                name = f'{prefix}_{k}'
                print('    ', name.ljust(33), '= ', f'0x{j:04X}', sep='')
                j += 1



# fmt: off

import asyncio
from itertools import islice

from agentica_internal.core.print import hprint
from agentica_internal.testing.examples import ITERATOR_FNS
from agentica_internal.warpc.worlds.debug_world import DebugWorld


def realize(it) -> list:
    return list(islice(it, 32))


async def verify_iterator_content(iter_fn):

    # verify the *content* of iterators
    pair = DebugWorld.connected_pair(logging=False)
    async with pair as B:
        content_a = realize(iter_fn())

        iter_obj = iter_fn()
        iter_obj_b = await B(iter_obj)
        content_b = realize(iter_obj_b)

        assert content_a == content_b, f"Content not equal: {content_a} != {content_b}"

    return pair


async def verify_iterators():
    for iter_fn in ITERATOR_FNS:
        hprint(iter_fn())
        await verify_iterator_content(iter_fn)


def collect_iterator_events(events: list):
    for iter_fn in ITERATOR_FNS:
        pair = asyncio.run(verify_iterator_content(iter_fn))
        pair.collect_events(events)
        del pair


async def verify_generator():
    await verify_iterator_content(lambda: gen_fn())


async def verify_re_finditer():
    import re

    # verify the *content* of iterators
    pair = DebugWorld.connected_pair(logging=False)
    async with pair as B:

        patt = re.compile(" X+ ")
        iter_fn = lambda: patt.finditer(" X XX XXX ")

        list_a = list(iter_fn())
        list_b = await B(iter_fn())

        assert isinstance(list_b, list)
        assert isinstance(list_b[0], re.Match)
        assert repr(list_a) == repr(list_b)

    return pair


if __name__ == '__main__':
    from agentica_internal.testing.examples.iterators import gen_fn
    asyncio.run(verify_generator())
    asyncio.run(verify_iterators())
    asyncio.run(verify_re_finditer())



# fmt: off

import re

from os import getenv
from sys import __stderr__ as STDERR
# from .log_streams import write_err

from .log_env import get_log_tags
from .log_groups import *

__all__ = [
    'should_log_cls',
    'should_log_tag',
    'cls_log_tags',
]

###############################################################################

def tag_match(tags_set: set[str], tags_str: str):
    if tags_str == 'all':
        return True
    if tags_str == '1' and '_' not in tags_set:
        return True
    tags_set = set(t.lower() for t in tags_set)
    tags_str = GROUP_RE.sub(lambda g: GROUP_EXPAND(g.group(0)), tags_str)
    if '+' not in tags_str:
        return tags_str in tags_set
    if not any(tag in tags_str for tag in tags_set):
        return False
    return bool(set(tags_str.split('+')).intersection(tags_set))

GROUP_RE = re.compile(rf'\b({'|'.join(GROUP_NAMES)})\b')

###############################################################################


###############################################################################

def should_log_tag(default: bool, tags: set[str]) -> bool:
    curr = get_log_tags()
    if curr == 'ALL':
        return True
    if curr == '1' and '_' not in tags:
        flag = True
    elif curr:
        flag = tag_match(tags, curr.lower())
    else:
        flag = bool(default)
    if getenv('AGENTICA_DEBUG_TAGS') == '1':
        STDERR.write(f"should_log_tag({default}, {tags}; env={curr!r}) => {flag}\n")
    return flag


def should_log_cls(flag: bool, cls: type):
    return should_log_tag(flag, cls_log_tags(cls))

###############################################################################

def cls_log_tags(cls: type) -> set[str]:
    if tags := CLS_LOG_TAGS.get(cls):
        return tags
    name = cls.__name__
    tagset = {name.lower()}
    extra = getattr(cls, 'LOG_TAGS', ())
    if isinstance(extra, str):
        extra = (extra,)
    tagset |= set(e.lower() for e in extra)
    for base in cls.__bases__:
        if not base.__flags__ & 256:
            tagset |= set(cls_log_tags(base))
    CLS_LOG_TAGS[cls] = tags = set(tagset)
    return tags

CLS_LOG_TAGS: dict[type, set[str]] = {}



from agentica_internal.repl.repl import *


def verify_return_statement():
    repl = BaseRepl()
    ev = repl.run_code('''
    return 1 + 1
    ''')
    assert not ev.error
    assert ev.return_value == 2


def verify_return_statement_nesting():
    repl = BaseRepl()
    # a return in a function should not cause a global return
    ev = repl.run_code('''
    def foo():
        return 1 + 1
    for i in range(5,10):
        return foo() + i
    ''')
    assert not ev.error
    assert ev.return_value == 7


def verify_return_variable():
    repl = BaseRepl()
    repl.options.return_var = 'result'
    ev = repl.run_code('''
    result = 9
    ''')
    assert not ev.error
    assert ev.return_value == 9


def verify_return_variable_nesting():
    repl = BaseRepl()
    repl.options.return_var = 'result'
    # setting a local result in a function should not cause a return
    # TODO: recognise 'global result' as an exception to this
    ev = repl.run_code('''
    def foo():
        result = 9
    foo()
    ''')
    assert not ev.error
    assert not ev.has_return_value


def verify_does_not_catch_returnexit():
    repl = BaseRepl()
    ev = repl.run_code('''
    try:
        return 'good'
    except:
        return 'bad'
    ''')
    assert ev.return_value == 'good'

    ev = repl.run_code('''
    try:
        return 'good'
    finally:
        pass
    ''')
    assert ev.return_value == 'good'

    ev = repl.run_code('''
    try:
        return 'good'
    except SystemExit:
        pass
    ''')
    assert ev.return_value == 'good'

    ev = repl.run_code('''
    try:
        try:    
            return 'good'
        except:
            return 'bad'
    except:
        return 'bad'
    ''')
    assert ev.return_value == 'good'


if __name__ == '__main__':
    verify_return_statement()
    verify_return_statement_nesting()
    verify_return_variable()
    verify_return_variable_nesting()
    verify_does_not_catch_returnexit()



# fmt: off

from ..shared import *

__all__ = [
    'hetero_dict_obj',
    'hetero_tuple_obj',
    'homo_tuple_obj',
    'homo_set_obj',
    'homo_list_obj',
    'homo_dict_obj',
    'HOMO_OBJS',
    'HETERO_OBJS',
    'CONTAINER_OBJS',
]

homo_tuple_obj = tuple(_strs)
homo_set_obj = set(_strs)
homo_list_obj = list(_strs)
homo_dict_obj = dict.fromkeys(_strs, _int1)

HOMO_OBJS: list[object] = [
    homo_tuple_obj,
    homo_set_obj,
    homo_list_obj,
    homo_dict_obj,
]

hetero_dict_obj = {'a': _a, 'b': _b, 'c': _c}
hetero_tuple_obj = (_a, _b, _c)

HETERO_OBJS: list[object] = [
    hetero_dict_obj,
    hetero_tuple_obj,
]

CONTAINER_OBJS = HOMO_OBJS + HETERO_OBJS



# fmt: off

from .__ import *
from .base import Request, ResultCoro


__all__ = [
    'ResourceRequest',
    'ResourceNew',
    'ResourceCallRequest',
    'ResourceCallFunction',
    'ResourceCallMethod',
    'ResourceCallSystemMethod',
    'ResourceAttrRequest',
    'ResourceGetAttr',
    'ResourceHasAttr',
    'ResourceDelAttr',
    'ResourceSetAttr',
    'GENERIC_RESOURCE_ERROR'
]


################################################################################

class ResourceRequest(Request, ABC):
    """
    ABC for a fully decoded request on a resource.

    It can be executed via `.execute` or `.execute_async` methods.

    It can be encoded to a `ResourceRequestMsg` via `.encode`.

    Instances of some subclasses can have the `async_` slot set, which communicates
    the intent for how they should be executed. This slot is only present on those
    requests for which this makes sense, preventing misuse.
    """

    __slots__ = __match_args__ = ()

    ############################################################################

    @abstractmethod
    def encode(self, codec: 'EncoderP') -> 'ResourceRequestMsg':
        """
        Encode a locally expressed resource request (in which typically the first
        argument is a remote resource virtualized locally) into a wire message.
        """
        ...

    ############################################################################

    def execute(self) -> Result:
        """
        Executes a decoded remote resource request immediately, wrapping the
        result in a Result.This happens through `__execute__`, which actually
        does the operation without worrying about catching exceptions.
        """
        try:
            value = self.__execute__()
            return Result.good(value)
        except BaseException as e:
            return Result.bad(e)

    @abstractmethod
    def __execute__(self) -> Any: ...

    ############################################################################

    def execute_async(self) -> ResultCoro:
        """
        Executes a decoded remote request, but not immediately: defers the
        creation of the Result of the execution via a coroutine function.

        This coroutine function will run the underlying exec, assume it *returns*
        an awaitable result (e.g. a coroutine object), await this, and then wrap
        the output in a Result.

        This is called by `Frame.exec_incoming_request_task`.
        """
        if self.is_sync:
            async def execute() -> Result:
                return self.execute()
        else:
            async def execute() -> Result:
                try:
                    awaitable = self.__execute__()
                    value = await awaitable
                    return Result.good(value)
                except BaseException as e:
                    return Result.bad(e)

        # ensure the coroutine object has a name that matches the underlying
        # resource
        coro = execute()
        coro.__qualname__ = coro.__name__ = self.async_name()
        return coro

    ############################################################################

    def hook_key(self) -> A.Hashable: ...


################################################################################

class ResourceNew(ResourceRequest):
    __slots__ = __match_args__ = 'cls', 'pos', 'key'

    cls: type
    pos: ArgsT
    key: KwargsT

    def __init__(self, cls: type, pos: ArgsT = (), key: KwargsT = {}) -> None:
        self.cls = cls
        self.pos = pos
        self.key = key

    def __execute__(self):
        return self.cls(*self.pos, **self.key)

    def encode(self, enc) -> 'ResourceNewMsg':
        from ..msg.rpc_request_resource import ResourceNewMsg
        return ResourceNewMsg(
            enc.enc_remote_resource(self.cls),
            enc.enc_args(self.pos),
            enc.enc_kwargs(self.key)
        )

    def hook_key(self):
        return 'new'


################################################################################

class ResourceCallRequest(ResourceRequest, ABC):
    __slots__ = __match_args__ = 'pos', 'key', 'async_mode'


################################################################################

class ResourceCallFunction(ResourceCallRequest):
    __slots__ = __match_args__ = 'fun', 'pos', 'key', 'async_mode'

    fun: Callable
    pos: ArgsT
    key: KwargsT

    def __init__(self, fun: Callable, pos: ArgsT = (), key: KwargsT = {}) -> None:
        self.fun = fun
        self.pos = pos
        self.key = key

    def __execute__(self):
        return self.fun(*self.pos, **self.key)

    def async_name(self):
        fun_name = getattr(self.fun, '__qualname__')
        return fun_name

    def encode(self, enc) -> 'ResourceCallFunctionMsg':
        from ..msg.rpc_request_resource import ResourceCallFunctionMsg
        return ResourceCallFunctionMsg(
            enc.enc_remote_resource(self.fun),
            enc.enc_args(self.pos),
            enc.enc_kwargs(self.key)
        )

    def hook_key(self):
        return 'call'


################################################################################

class ResourceCallMethod(ResourceCallRequest):
    __slots__ = __match_args__ = 'obj', 'mth', 'pos', 'key', 'async_mode'

    obj: object
    mth: str
    pos: ArgsT
    key: KwargsT

    def __init__(self, obj: object, meth: str, pos: ArgsT = (), key: KwargsT = {}) -> None:
        self.obj = obj
        self.mth = meth
        self.pos = pos
        self.key = key

    def __execute__(self):
        fun = getattr(self.obj, self.mth)
        return fun(*self.pos, **self.key)

    def async_name(self):
        cls_name = type(self).__qualname__
        mth_name = self.mth
        return f'{cls_name}.{mth_name}'

    def encode(self, enc) -> 'ResourceCallMethodMsg':
        from ..msg.rpc_request_resource import ResourceCallMethodMsg
        return ResourceCallMethodMsg(
            enc.enc_remote_resource(self.obj),
            self.mth,
            enc.enc_args(self.pos),
            enc.enc_kwargs(self.key)
        )

    def hook_key(self):
        return 'call_method', self.mth


################################################################################

class ResourceCallSystemMethod(ResourceCallRequest):
    __slots__ = __match_args__ = 'fun', 'obj', 'async_mode'

    obj: object
    fun: Callable

    def __init__(self, obj: object, fun: Callable):
        self.obj = obj
        self.fun = fun

    def async_name(self):
        cls_name = type(self).__qualname__
        fun_name = getattr(self.fun, '__name__')
        return f'{fun_name}(<{cls_name}>)'

    def __execute__(self):
        return self.fun(self.obj)

    def encode(self, enc) -> 'ResourceCallSystemMethodMsg':
        from ..msg.rpc_request_resource import ResourceCallSystemMethodMsg
        return ResourceCallSystemMethodMsg(
            enc.enc_remote_resource(self.obj),
            enc.enc_system_resource(self.fun),
        )

    def hook_key(self):
        return self.fun



################################################################################

class ResourceAttrRequest(ResourceRequest, ABC):
    __slots__ = __match_args__ = 'obj', 'attr',

    obj:  object
    attr: str


################################################################################

class ResourceHasAttr(ResourceAttrRequest):
    __slots__ = __match_args__ = 'obj', 'attr',

    def __init__(self, obj: object, attr: str):
        self.obj = obj
        self.attr = attr

    def __execute__(self):
        return hasattr(self.obj, self.attr)

    def encode(self, enc) -> 'ResourceHasAttrMsg':
        from ..msg.rpc_request_resource import ResourceHasAttrMsg
        return ResourceHasAttrMsg(
            enc.enc_remote_resource(self.obj),
            self.attr,
        )

    def hook_key(self):
        return hasattr


################################################################################

class ResourceGetAttr(ResourceAttrRequest):
    __slots__ = __match_args__ = 'obj', 'attr',

    def __init__(self, obj: object, attr: str):
        if type(attr) is not str:
            bad_attr(attr)
        self.obj = obj
        self.attr = attr

    def __execute__(self):
        return getattr(self.obj, self.attr)

    def encode(self, enc) -> 'ResourceGetAttrMsg':
        from ..msg.rpc_request_resource import ResourceGetAttrMsg
        return ResourceGetAttrMsg(
            enc.enc_remote_resource(self.obj),
            self.attr,
        )

    def hook_key(self):
        return getattr


################################################################################

class ResourceDelAttr(ResourceAttrRequest):
    __slots__ = __match_args__ = 'obj', 'attr',

    def __init__(self, obj: object, attr: str):
        if type(attr) is not str:
            bad_attr(attr)
        self.obj = obj
        self.attr = attr

    def __execute__(self):
        delattr(self.obj, self.attr)

    def encode(self, enc) -> 'ResourceDelAttrMsg':
        from ..msg.rpc_request_resource import ResourceDelAttrMsg
        return ResourceDelAttrMsg(
            enc.enc_remote_resource(self.obj),
            self.attr,
        )

    def hook_key(self):
        return delattr


################################################################################

class ResourceSetAttr(ResourceAttrRequest):
    __slots__ = __match_args__ = 'obj', 'attr', 'val',

    obj: object
    attr: str
    val: Any

    def __init__(self, obj: object, attr: str, val: Any):
        if type(attr) is not str:
            bad_attr(attr)
        self.obj = obj
        self.attr = attr
        self.val = val

    def __execute__(self):
        setattr(self.obj, self.attr, self.val)

    def encode(self, enc) -> 'ResourceSetAttrMsg':
        from ..msg.rpc_request_resource import ResourceSetAttrMsg
        return ResourceSetAttrMsg(
            enc.enc_remote_resource(self.obj),
            self.attr,
            enc.enc_any(self.val),
        )

    def hook_key(self):
        return setattr


################################################################################


def bad_attr(attr) -> NoReturn:
    raise TypeError(f'attribute name must be a string, not {f_object(attr)}')


GENERIC_RESOURCE_ERROR = Result.bad(RuntimeError('resource no longer available'))



# fmt: off

import logging
import uuid

from enum import Enum, EnumMeta, IntEnum, StrEnum
from types import MethodType

from ..core.log import LogBase
from ..cpython.iters import *
from .__ import *
from .hooks import *
from .attrs import FUTURE_ID, WARP_AS, CLASS_WARP_AS
from .kinds import *
from .msg.all import *
from .msg.term import COMPLEX_ENCODERS, CONSTANT_ENCODERS, SIMPLE_ENCODERS
from .repl import *
from .request.all import *
from .resource.all import *
from .resource.logging import *
from .system import FORBIDDEN_IDS, LRID_TO_SRID, SRID_TO_RSRC

__all__ = [
    'WorldP',
    'Frame',
    'ExecRemoteFn',
    'ResponseMsgFn',
]


################################################################################

if TYPE_CHECKING:
    from asyncio import AbstractEventLoop, Task

################################################################################

type ResponseMsgFn = Callable[[FramedResponseMsg], Any]
type ExecRemoteFn = Callable[[FramedRequestMsg], FramedResponseMsg]

################################################################################

class WorldP(Protocol):
    """WorldP is a protocol describing the API surface of World that is relevant
    for how virtual resources are handled by Frame on behalf of World."""

    def get_log_name(self) -> str: ...
    def get_world_id(self) -> WorldID: ...
    def get_repl(self) -> ReplP | None: ...

    def get_frame(self, fid: FrameID) -> 'Frame | None': ...
    def delete_frame(self, frame: 'Frame'): ...

    def new_message_id(self) -> MessageID: ...

    def register_future(self, future: FutureT, /) -> None: ...
    def unregister_future(self, future: FutureT, /) -> None: ...
    def future_from_id(self, future_id: FutureID, /) -> FutureT: ...

    def resource_post_encode(self, frame: 'Frame', data: ResourceData, grid: GlobalRID) -> None: ...
    def resource_pre_decode(self, frame: 'Frame', data: ResourceData, grid: GlobalRID) -> None: ...
    def resource_post_decode(self, resource: ResourceT, handle: ResourceHandle) -> None: ...

    def send_event(self, msg: EventMsg) -> None: ...
    def execute_outgoing_request_sync(self, msg: FramedRequestMsg) -> FramedResponseMsg: ...
    def execute_outgoing_request_coro(self, msg: FramedRequestMsg) -> Awaitable[FramedResponseMsg]: ...
    def execute_outgoing_request_future(self, msg: FramedRequestMsg) -> Awaitable[FramedResponseMsg]: ...

    def will_hook(self, request: ResourceRequest) -> bool: ...
    def get_resource_request_hook_result(self, remote: Callable[[], Result], handle: ResourceHandle, request: ResourceRequest) -> Result: ...


################################################################################

class Frame(LogBase, CodecP):
    __slots__ = (
        'wid', 'fid', 'pid', 'key', 'dep', 'man',
        'lrid_to_rmsg', 'grid_to_rsrc',
        'outgoing_defs', 'incoming_defs',
        'local_lrids', 'remote_lrids'
    )

    wid: WorldID  # our world id
    fid: FrameID  # our frame id
    pid: FrameID  # parent frame id

    # used to look up a world and frame unambiguously
    key: FrameKey

    depth: int
    world: WorldP

    outgoing_defs:  list[DefinitionMsg]             # definitions we have sent
    incoming_defs:  dict[GlobalRID, DefinitionMsg]  # definitions we have received via .defs
    lrid_to_rmsg:   dict[LocalRID, UserResourceMsg]  # local OR remote resources sent or received
    grid_to_rsrc:   dict[GlobalRID, ResourceT]
    local_lrids:    set[LocalRID]
    remote_lrids:   set[LocalRID]

    __enc_stack:    list[ResourceMsg]
    __dec_stack:    list[GlobalRID]
    __obj_stack:    list[object]  # for debugging

    ############################################################################

    def extend(self, other: Self) -> None:
        self.lrid_to_rmsg.update(other.lrid_to_rmsg)
        self.grid_to_rsrc.update(other.grid_to_rsrc)

    ############################################################################

    def __init__(self, fid: FrameID, pid: FrameID, wid: WorldID):
        self.wid = wid
        self.fid = fid
        self.pid = pid
        self.key = wid, fid
        super().__init__(id_name=f_id(fid))

    ############################################################################

    def init_root(self, world: WorldP):
        self.world = world
        self.lrid_to_rmsg = {}
        self.grid_to_rsrc = {}
        self.local_lrids = set()
        self.remote_lrids = set()
        self.outgoing_defs = []
        self.incoming_defs = {}
        self.log_name = world.get_log_name() + '.' + self.log_name
        self.__enc_stack = []
        self.__dec_stack = []
        self.__obj_stack = []

    def cloned_root(self, world: WorldP) -> Self:
        cloned = Frame(self.fid, self.pid, self.wid)
        cloned.world = world
        cloned.lrid_to_rmsg = self.lrid_to_rmsg.copy()
        cloned.grid_to_rsrc = self.grid_to_rsrc.copy()
        cloned.local_lrids = self.local_lrids.copy()
        cloned.remote_lrids = self.remote_lrids.copy()
        cloned.outgoing_defs = self.outgoing_defs.copy()
        cloned.incoming_defs = {}
        cloned.log_name = world.get_log_name() + '.' + self.log_name
        cloned.__enc_stack = []
        cloned.__dec_stack = []
        cloned.__obj_stack = []
        return cloned

    def inherit(self, fid: FrameID = 0) -> 'Frame':
        return self

    def delete(self) -> None:
        pass

    ############################################################################

    def exec_incoming_request(
            self,
            request_msg: FramedRequestMsg,
            send_fn:     ResponseMsgFn | None,
            exit_fn:     ResponseMsgFn
    ) -> FramedResponseMsg:
        """
        Decodes an incoming `FramedRequestMsg`, execute it, then immediately
        encode the result and send it via a `FramedResponseMsg`.

        This is called by `SDKWorld.handle_incoming_message` and
        `AgentWorld.run_until_mid`.
        """
        request = request_msg.decode_request(self)
        result = request.execute()
        response_msg = request_msg.encode_response(self, result)
        exit_fn(response_msg)
        # TODO: just have a BaseWorld.send_later that either sends immediately
        if send_fn:
            assert send_fn(response_msg) is None  # send_fn cannot be an async function
        return response_msg

    def exec_incoming_request_task(
            self,
            request_msg:  FramedRequestMsg,
            loop:        'AbstractEventLoop',
            send_fn:      ResponseMsgFn | None,
            exit_fn:      ResponseMsgFn
    ) -> 'Task[FramedResponseMsg]':
        """
        Decodes an incoming FramedRequestMsg, then creates an asyncio task to
        execute it in via a coroutine.

        This is called by worlds when they receive an incoming
        `FramedRequestMsg` and this message is marked as needing to execute
        asynchronously (via `async_mode = 'coro' | 'future'`).

        This method lives in `Frame` because it natively uses this frame's
        state to decode the incoming  `FramedRequestMsg` and encode the outgoing
        `FramedResponseMsg`.

        This is called by `SDKWorld.handle_incoming_message` and
        `AgentWorld.run_until_mid`.
        """

        request = request_msg.decode_request(self)
        result_coro = request.execute_async()

        async def respond():
            result = await result_coro
            response_msg = request_msg.encode_response(self, result)
            exit_fn(response_msg)
            if send_fn:
                obj = send_fn(response_msg)
                # if callback was an async function, (e.g. send_xxx), wait for it to complete
                if isinstance(obj, Awaitable):
                    await obj
            return response_msg

        respond_coro = respond()
        task = loop.create_task(respond_coro)

        return task

    ############################################################################

    def get_user_resource(self, grid: GlobalRID) -> ResourceT:
        resource = self.grid_to_rsrc.get(grid)
        if resource is not None:
            return resource
        def_msg = self.incoming_defs.get(grid)
        if def_msg is not None:
            return self.decode_incoming_definition(grid, def_msg)
        self.missing_user_resource(grid)

    def missing_user_resource(self, grid: GlobalRID) -> NoReturn:
        grid_ = f_grid(grid)
        name = self.log_name
        if LOG_DECR:
            P.hdiv()
            P.nprint(ICON_C1, 'missing virtual resource', grid_, 'in', name)
        P.eprint(ICON_C1, 'missing virtual resource', grid_, 'in', name)
        problem = f"resource with grid={grid_} is not defined.\n"
        self.raise_dec_err(problem)

    ############################################################################

    def decode_incoming_definition(
            self,
            grid: GlobalRID,
            def_msg: DefinitionMsg,
        ) -> ResourceT:

        obj_stack = self.__obj_stack
        dec_stack = self.__dec_stack

        if grid in dec_stack:
            fallback_resource = self.handle_decode_recursion(grid, def_msg.data)
            if fallback_resource is not None:
                return fallback_resource

        # decode the data field, e.g. ClassDataMsg into ClassData
        try:
            dec_stack.append(grid)
            data = def_msg.data.decode(self)
        finally:
            dec_stack.pop()

        if log := bool(LOG_DECR):
            P.hdiv()
            P.nprint(ICON_C0, 'creating virtual resource in', self, 'from data message:\n')
            def_msg.data.pprint()

        # world can mess with the data if it wants
        self.world.resource_pre_decode(self, data, grid)

        # create a virtual resource that uses us to service requests
        handle = self.new_resource_handle(grid)
        try:
            obj_stack.append(data)
            remote_resource = data.create_resource(handle)
        finally:
            obj_stack.pop()

        # world can inspect the newly decoded resource if it wants
        self.world.resource_post_decode(remote_resource, handle)

        # bind this remote resource and the message to send it as under its ids
        self.grid_to_rsrc[grid] = remote_resource
        if not data.MIGHT_ALIAS:
            lrid = id(remote_resource)
            self.lrid_to_rmsg[lrid] = UserResourceMsg(grid)  # so we can send it back
            self.remote_lrids.add(lrid)

        if log:
            P.nprint()
            P.nprint(ICON_C1, 'created', remote_resource, 'in', self, 'with handle', handle)
            P.hdiv()

        return remote_resource

    def handle_decode_recursion(self, grid: GlobalRID, data: ResourceDataMsg) -> ResourceT:
        if isinstance(data, TypeAliasDataMsg):
            alias_name = data.name
            alias_module = data.module
            return T.ForwardRef(alias_name, is_argument=False, module=alias_module)
        elif isinstance(data, ClassDataMsg):
            alias_name = data.qname
            alias_module = data.module
            return T.ForwardRef(alias_name, is_argument=False, module=alias_module, is_class=True)
        elif isinstance(data, TypeData):
            return Any
        else:
            self.raise_dec_err(f'cannot decode grid {f_grid(grid)} for recursive resource: {data}')

    # --------------------------------------------------------------------------

    def encode_outgoing_resource(self, lrid: LocalRID, resource: ResourceT) -> ResourceMsg:

        enc_stack = self.__enc_stack
        obj_stack = self.__obj_stack

        # choose the kind of data class
        data_cls = choose_resource_class(resource)

        # make a fresh GlobalRID
        grid = self.new_global_resource_id(lrid)

        if log := bool(LOG_ENCR):
            P.hdiv()
            P.nprint(ICON_C0, 'describing real resource', f_object_id(resource), 'in', self)
            P.nprint()

        # bind this local resource and the message to send it as under its ids
        resource_msg = UserResourceMsg(grid)
        self.lrid_to_rmsg[lrid] = resource_msg
        self.grid_to_rsrc[grid] = resource
        self.local_lrids.add(lrid)

        # make an ResourceData instance (e.g. ClassData)

        try:
            obj_stack.append(resource)
            data = data_cls.describe_resource(resource)
        except E.WarpEncodingForbiddenError:
            if log:
                P.nprint(ICON_C1, 'forbidden error while encoding:', resource)
                P.hdiv()
            return data_cls.forbidden_msg()
        finally:
            obj_stack.pop()

        # let the world mess with the data if it wants
        self.world.resource_post_encode(self, data, grid)

        if log:
            data.pprint()
            P.nprint()
            P.nprint(ICON_C1, 'described resource; grid =', f_grid(grid))
            P.hdiv()

        # encode the ResourceData to a ResourceDataMsg

        try:
            obj_stack.append(resource)
            enc_stack.append(resource_msg)
            data_msg = data.encode(self)
        finally:
            obj_stack.pop()
            enc_stack.pop()

        # wrap the ResourceDataMsg in a DefinitionMsg
        def_msg = DefinitionMsg(grid, data_msg)

        # add this definition for benefit of enc_context
        self.outgoing_defs.append(def_msg)

        if log:
            self.log('encoded', f_grid(grid), 'from', resource)

        return resource_msg

    def new_global_resource_id(self, ptr: Ptr) -> GlobalRID:
        return self.wid, self.fid, ptr

    def new_resource_handle(self, grid: GlobalRID) -> ResourceHandle:
        handle = ResourceHandle()
        handle.grid = grid
        handle.fkey = self.key
        handle.hdlr = self.handle_virtual_resource_request
        return handle

    # --------------------------------------------------------------------------

    def handle_virtual_resource_request(self, origin: ResourceHandle, req: ResourceRequest) -> Any:

        if log := bool(LOG_VIRT):
            P.hdiv()
            P.nprint(ICON_I, 'VRR of', origin, 'in', self, 'of', 'in', self.world)
            req.pprint()
            P.hdiv()

        result: Result

        if flags.VIRTUAL_RESOURCE_REQUEST_HOOKS and self.world.will_hook(req):
            P.nprint(ICON_I, 'running hook for', req.hook_key()) if log else None
            remote_fn = partial(self.remote_resource_request_result, origin, req)
            result = self.world.get_resource_request_hook_result(remote_fn, origin, req)
        else:
            result = self.remote_resource_request_result(origin, req)

        if log:
            P.nprint(ICON_O, result)

        return result.realize()

    ############################################################################

    # this executes an outgoing request and returns the result
    # if this is an .as_future or .as_coro request, we can return immediately,
    # otherwise, we use world.execute_outgoing_request_sync(msg) to block until the
    # eventual FramedResponseMsg is received
    def remote_resource_request_result(self, origin: ResourceHandle, request: ResourceRequest) -> Result:
        """
        This executes an outgoing request and returns the result as a `Result`.

        If this is an asynchronous request (`.is_async` are True), we will return
        a coroutine / future immediately.

        Otherwise, we use  `world.execute_outgoing_request_sync(msg)` to  block
        until the eventual FramedResponseMsg is received.
        """

        log = bool(LOG_VIRT)
        if origin.fkey is not self.key:
            P.nprint(ICON_E, f'origin key mismatch:', origin.fkey, '!=', self.key) if log else None
            return GENERIC_RESOURCE_ERROR

        mid = self.world.new_message_id()
        msg = FramedRequestMsg.encode_request(self, mid, self.fid, request)

        if request.as_future:
            P.nprint(ICON_A, 'sending future-based request') if log else None

            # this will create an `awaiting` future for the FramedResponseMsg
            response_future = self.world.execute_outgoing_request_future(msg)
            P.nprint(ICON_A, 'response future =', response_future) if log else None

            P.nprint(ICON_A, 'created derived future for result') if log else None
            # this will create a derived future that decodes this awaiting future
            result_future = self.world.future_from_id(msg.mid)

            P.nprint(ICON_M, 'returning result future =', result_future) if log else None
            return Result.good(result_future)

        elif request.as_coro:

            async_name = request.async_name()
            log_name = f'<async {async_name} #{mid}>'

            async def dummy_fn():
                P.nprint(ICON_A, log_name, 'coroutine started') if log else None
                awaitable = self.world.execute_outgoing_request_coro(msg)
                P.nprint(ICON_A, log_name, 'awaiting response') if log else None
                assert isinstance(awaitable, Awaitable)
                response_msg = await awaitable
                final_result = response_msg.decode_response(self)
                P.nprint(ICON_A, log_name, 'got response', final_result) if log else None
                return final_result.realize()

            dummy_fn.__name__ = dummy_fn.__qualname__ = async_name
            dummy_fn.__module__ = 'virtual'
            dummy_fn.__code__ = dummy_fn.__code__.replace(
                co_name=async_name,
                co_qualname=async_name,
                co_filename='virtual'
            )

            P.nprint(ICON_M, log_name, 'returning coroutine immediately') if log else None
            coro = dummy_fn()
            return Result.good(coro)

        response = self.world.execute_outgoing_request_sync(msg)

        # if ctx:
        #     ctx.info('received response:', response)

        if not isinstance(response, FramedResponseMsg):
            cause = f'expected FramedResponseMsg, not {f_object_id(response)}'
            raise E.WarpProtocolError(cause)

        result = response.decode_response(self)

        # if ctx:
        #     ctx.info('decoded result:', result)

        return result

    ############################################################################

    def enc_any(self, term: TermT) -> TermMsg:

        msg = self._enc_val(term)
        if msg is not None:
            return msg

        cls = type(term)

        if flags.TYPE_ERASE_ENUMS and type(cls) is EnumMeta:
            if issubclass(term, IntEnum):
                val_cls = int
            elif issubclass(term, StrEnum):
                val_cls = str
            else:
                val_cls = object
            return self.enc_class(val_cls)

        if flags.REALIZE_SYSTEM_ITERATORS:
            if cls.__flags__ & 256 and cls in ITER_TYPES:
                realized = list(term)
                return self._enc_val(realized)

        if flags.VIRTUAL_LAMBDAS:
            if cls is FunctionType and term.__name__ == '<lambda>':
                try:
                    return LambdaMsg.encode_lambda(self, term)
                except:
                    pass

        return self.enc_resource(term)

    # --------------------------------------------------------------------------

    def enc_exception(self, exc: BaseException) -> ExceptionMsg:
        if not isinstance(exc, BaseException):
            self.raise_enc_err(f"not an exception: {f_object_id(exc)}")
        return ExceptionMsg.encode_compound(exc, self)

    def enc_value(self, val: ValueT) -> TermPassByValMsg:
        if msg := self._enc_val(val):
            return msg
        self.raise_enc_err(f"{f_object_id(val)} is not a pass-by-value term")

    def _enc_val(self, val: ValueT) -> TermPassByValMsg | None:

        cls = type(val)

        if msg := const_fn(cls):
            return msg

        if fn := atom_fn(cls):
            return fn(val)

        if fn := compound_fn(cls):
            return fn(val, self)

        if isinstance(val, BaseException):
            return ExceptionMsg.encode_compound(val, self)

        if isinstance(val, Enum):
            return EnumMemberMsg.encode_enum(val, self)

        return None

    # --------------------------------------------------------------------------

    def enc_object(self, obj: ObjectT) -> ResourceMsg:
        if not is_object_t(obj):
            raise E.WarpEncodingError(f"not an object: {f_object_id(obj)}")
        return self.enc_resource(obj)

    def enc_class(self, cls: ClassT) -> ResourceMsg:
        if not is_class_t(cls):
            self.raise_enc_err(f"not a class: {f_object_id(cls)}")
        return self.enc_resource(cls)

    def enc_type(self, cls: TypeT) -> TermMsg:
        try:
            return self.enc_any(cls)
        except E.WarpEncodingError:
            return self.enc_resource(Any)

    def enc_function(self, fun: FunctionT) -> ResourceMsg:
        if not is_function_t(fun):
            self.raise_enc_err(f"not a function: {f_object_id(fun)}")
        return self.enc_resource(fun)

    def enc_module(self, mod: ModuleT) -> ResourceMsg:
        if not is_module_t(mod):
            self.raise_enc_err(f"not a module: {f_object_id(mod)}")
        return self.enc_resource(mod)

    def enc_future(self, fut: FutureT) -> ResourceMsg:
        if not is_future_t(fut):
            self.raise_enc_err(f"not a future: {f_object_id(fut)}")
        return self.enc_resource(fut)

    def enc_resource(self, resource: ResourceT) -> ResourceMsg:
        lrid = id(resource)

        # is it a known, system resource?
        if sid := LRID_TO_SRID.get(lrid):
            return SystemResourceMsg(sid)

        # have we encoded this resource before?
        if msg := self.lrid_to_rmsg.get(lrid):
            return msg

        if flags.RESPECT_WARP_AS:
            cls = type(resource)
            if cls is type or isinstance(resource, type):
                if warp_as := getattr(resource, CLASS_WARP_AS, None):
                    if type(warp_as) is MethodType:
                        self.log("class resource will warp as something else via", warp_as)
                        as_class = warp_as()
                        self.log("warping as", as_class)
                        return self.enc_class(as_class)
            else:
                if warp_as := getattr(cls, WARP_AS, None):
                    if callable(warp_as):
                        self.log("resource will warp as something else via", warp_as)
                        as_resource = warp_as(resource)
                        self.log("warping as", as_resource)
                        return self.enc_any(as_resource)  # type: ignore

        handle = get_handle(resource)
        if handle is not None:
            f_resource = f_object_id(resource)
            self.log("existing", resource, "with handle", handle, "not in cache")
            msg = f"existing {f_resource} with handle {handle} not in cache"
            self.raise_enc_err(msg)

        return self.encode_outgoing_resource(lrid, resource)

    def enc_local_resource(self, res: ResourceT) -> LocalResourceMsg:
        if has_handle(res):
            self.raise_enc_err(f'local resource expected, got {f_object_id(res)}')
        return self.enc_resource(res)

    def enc_remote_resource(self, res: ResourceT) -> RemoteResourceMsg:
        # NOTE: could also check remote_lrids, but this wouldn't work across frames
        if not has_handle(res):
            self.raise_enc_err(f'remote resource expected, got {f_object_id(res)}')
        return self.enc_resource(res)

    def enc_system_resource(self, res: ResourceT) -> SystemResourceMsg:
        sid = LRID_TO_SRID.get(id(res))
        if sid is None:
            self.raise_enc_err(f'system resource expected, got {f_object_id(res)}')
        return SystemResourceMsg(sid)

    # --------------------------------------------------------------------------

    def enc_sequence(self, seq: Iter) -> Tup[TermMsg]:
        return tuple(map(self.enc_any, seq))

    def enc_record(self, dct: dict) -> Rec[TermMsg]:
        return dict(zip(dct.keys(), map(self.enc_any, dct.values())))

    def enc_args(self, tup: ArgsT) -> ArgsMsg:
        enc_any = self.enc_any
        return tuple(enc_any(v) for v in tup if v is not ARG_DEFAULT)

    def enc_kwargs(self, rec: KwargsT) -> KwargsMsg:
        enc_any = self.enc_any
        return {k: enc_any(v) for k, v in rec.items() if v is not ARG_DEFAULT}

    def enc_annotations(self, rec: AnnotationsT) -> AnnotationsMsg:
        if not rec:
            return {}
        res = {}
        for k, v in rec.items():
            try:
                v_msg = self.enc_any(v)
                if isinstance(v_msg, SystemResourceMsg) and v_msg.sid in FORBIDDEN_IDS:
                    continue
                res[k] = v_msg
            except E.WarpEncodingError:
                pass
        return res

    def enc_methods(self, rec: MethodsT, /) -> 'MethodsMsg':
        enc_fun = self.enc_function
        dct = {}
        for k, v in rec.items():
            kind, func = unpack_method(v)
            dct[k] = kind, enc_fun(func)
        return dct

    def enc_owner(self) -> 'ResourceMsg | None':
        stack = self.__enc_stack
        return stack[-2] if len(stack) > 1 else None

    ############################################################################

    def dec_any(self, msg: TermMsg) -> TermT:
        if not isinstance(msg, TermMsg):
            self.raise_dec_err(f"{f_object_id(msg)} is not a TermMsg")
        return msg.decode(self)

    # --------------------------------------------------------------------------

    def dec_exception(self, msg: ExceptionMsg) -> BaseException:
        if not isinstance(msg, ExceptionMsg):
            self.raise_dec_err(f"{f_object_id(msg)} is not an ExceptionMsg")
        return msg.decode(self)

    def dec_value(self, msg: TermMsg) -> ValueT:
        if not isinstance(msg, TermPassByValMsg):
            self.raise_dec_err(f"{f_object_id(msg)} is not a TermPassByValMsg")
        return msg.decode(self)

    # --------------------------------------------------------------------------

    def dec_object(self, msg: ResourceMsg) -> ObjectT:
        obj = self.dec_resource(msg)
        is_object_t(obj) or self.raise_wrong_type(msg, obj, Kind.Object)
        return obj

    def dec_class(self, msg: ResourceMsg) -> ClassT:
        cls = self.dec_resource(msg)
        is_class_t(cls) or self.raise_wrong_type(msg, cls, Kind.Class)
        return cls

    def dec_function(self, msg: ResourceMsg) -> FunctionT:
        fun = self.dec_resource(msg)
        is_function_t(fun) or self.raise_wrong_type(msg, fun, Kind.Function)
        return fun

    def dec_module(self, msg: ResourceMsg) -> ModuleT:
        mod = self.dec_resource(msg)
        is_module_t(mod) or self.raise_wrong_type(msg, mod, Kind.Module)
        return mod

    def dec_future(self, msg: ResourceMsg) -> FutureT:
        mod = self.dec_resource(msg)
        is_future_t(mod) or self.raise_wrong_type(msg, mod, Kind.Future)
        return mod

    def dec_type(self, msg: TermMsg) -> TypeT:
        try:
            return self.dec_any(msg)
        except E.WarpDecodingError as err:
            P.nprint("decoding type anno", msg, "threw exception", err)
            return Any

    def dec_resource(self, msg: ResourceMsg) -> ResourceT:
        if isinstance(msg, SystemResourceMsg):
            return SRID_TO_RSRC[msg.sid]
        elif isinstance(msg, UserResourceMsg):
            return self.get_user_resource(msg.rid)
        else:
            self.raise_dec_err(f'not a resource msg: {f_object_id(msg)}')

    def dec_local_resource(self, msg: 'LocalResourceMsg') -> ResourceT:
        if isinstance(msg, SystemResourceMsg):
            return SRID_TO_RSRC[msg.sid]
        elif not isinstance(msg, UserResourceMsg):
            self.raise_dec_err(f'not an ResourceMsg: {msg}')
        rsrc = self.grid_to_rsrc[msg.rid]
        if id(rsrc) not in self.local_lrids:
            self.raise_dec_err(f'expected a local resource')
        return rsrc

    def dec_remote_resource(self, msg: 'RemoteResourceMsg') -> ResourceT:
        if not isinstance(msg, UserResourceMsg):
            self.raise_dec_err(f'not an UserResourceMsg: {msg}')
        rsrc = self.grid_to_rsrc[msg.rid]
        if id(rsrc) not in self.remote_lrids:
            self.raise_dec_err(f'expected a remote resource')
        return rsrc

    def dec_system_resource(self, msg: 'SystemResourceMsg') -> ResourceT:
        if not isinstance(msg, SystemResourceMsg):
            self.raise_dec_err(f'not an SystemResourceMsg: {msg}')
        return SRID_TO_RSRC[msg.sid]

    # --------------------------------------------------------------------------

    def dec_sequence(self, seq: Iter[TermMsg]) -> Tup[TermT]:
        return tuple(map(self.dec_any, seq))

    def dec_record(self, rec: Rec[TermMsg]) -> Rec[TermT]:
        if type(rec) is not dict:
            self.raise_dec_err(f"{f_object_id(rec)} is not dict (of TermMsgs)")
        return dict(zip(rec.keys(), map(self.dec_any, rec.values())))

    def dec_args(self, tup: ArgsMsg) -> ArgsT:
        dec_any = self.dec_any
        return tuple(v for m in tup if (v := dec_any(m)) is not ARG_DEFAULT)

    def dec_kwargs(self, rec: KwargsMsg) -> KwargsT:
        dec_any = self.dec_any
        return {k: v for k, m in rec.items() if (v := dec_any(m)) is not ARG_DEFAULT}

    def dec_annotations(self, rec: AnnotationsMsg) -> AnnotationsT:
        if type(rec) is not dict:
            self.raise_dec_err(f"{f_object_id(rec)} is not dict (of TermMsgs)")
        return dict(zip(rec.keys(), map(self.dec_type, rec.values())))

    def dec_methods(self, rec: MethodsMsg) -> MethodsT:
        if type(rec) is not dict:
            self.raise_dec_err(f"{f_object_id(rec)} is not dict (of MethodMsgs)")
        dct = {}
        for k, pair in rec.items():
            if type(pair) is tuple:
                kind, fun_msg = pair
                dct[k] = pack_method(kind, self.dec_function(fun_msg))
            else:
                assert isinstance(pair, ResourceMsg)
                dct[k] = self.dec_function(pair)
        return dct

    # --------------------------------------------------------------------------

    def get_repl(self) -> ReplP | None:
        return self.world.get_repl()

    def future_to_id(self, future: FutureT, /) -> FutureID:
        future_id: FutureID | None = getattr(future, FUTURE_ID, None)
        if future_id is None:
            future_id: str = uuid.uuid4().hex
            setattr(future, FUTURE_ID, future_id)
        else:
            assert isinstance(future_id, (int, str))
        self.world.register_future(future)
        return future_id

    def future_from_id(self, future_id: FutureID) -> FutureT:
        return self.world.future_from_id(future_id)

    ############################################################################

    def enc_context(self) -> 'FrameEncoderContext':
        return FrameEncoderContext(self)

    def dec_context(self, msgs: 'Tup[DefinitionMsg]') -> 'FrameDecoderContext':
        return FrameDecoderContext(self, msgs)

    ############################################################################

    def raise_dec_err(self, problem: str) -> NoReturn:
        raise self.add_note(E.WarpDecodingError(problem))

    def raise_wrong_type(self, msg: ResourceMsg, rsr: ResourceT, kind: Kind) -> NoReturn:
        problem = f"{msg} decoded to non-{kind}: {f_object_id(rsr)} in {self.log_name}"
        P.nprint(msg, f"decoded to non-{kind}", rsr, "in", self)
        raise E.WarpDecodingError(problem)

    def raise_enc_err(self, problem: str) -> NoReturn:
        raise self.add_note(E.WarpEncodingError(problem))

    def add_note(self, exc: BaseException) -> BaseException:
        exc.add_note(f"frame: {self.log_name}")
        if stack := self.__obj_stack:
            exc.add_note(f"stack: {fmt_obj_stack(stack)}")
        if id_set := self.local_lrids:
            exc.add_note(f"local resources: {fmt_id_set(id_set)}")
        if id_set := self.remote_lrids:
            exc.add_note(f"remote resources: {fmt_id_set(id_set)}")
        return exc

    ############################################################################

    def log_internal_error(self, msg: str, exc: BaseException) -> None:
        from ..core.color import RESET
        f_msg = f"INTERNAL WARP ERROR IN {self.name}: {msg}"
        frame_logger.error(f_msg, exc_info=exc)
        f_exception = D.fmt_exception(exc)
        P.eprint(f"{RESET}\n{ERROR_BAR}\n{f_msg}\n{f_exception}")


def enc_no_defs(enc, vals: Iter):
    return tuple(enc(v) for v in vals if v is not ARG_DEFAULT)


################################################################################

def choose_resource_class(resource: ResourceT) -> type[ResourceData]:
    if is_function_t(resource):
        return FunctionData
    elif is_class_t(resource):
        if type(resource) is EnumMeta:
            return EnumClassData
        return ClassData
    elif is_coroutine_t(resource):
        return CoroutineData
    elif is_type_t(resource):
        return TypeAliasData
    elif flags.VIRTUAL_MODULES and is_module_t(resource):
        return ModuleData
    elif flags.VIRTUAL_ITERATORS and is_iterator_t(resource):
        return IteratorData
    elif flags.VIRTUAL_FUTURES and is_future_t(resource):
        return FutureData
    else:
        return ObjectData

################################################################################

class FrameEncoderContext:

    outgoing: list[DefinitionMsg]
    start: int
    stop: int

    def __init__(self, frame: Frame):
        self.outgoing = frame.outgoing_defs

    def __enter__(self):
        self.start = len(self.outgoing)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop = len(self.outgoing)

    def enc_context_defs(self) -> Tup[DefinitionMsg]:
        defs = self.outgoing[self.start:self.stop]
        defs.reverse()
        return tuple(defs)


################################################################################

class FrameDecoderContext:

    incoming: dict[GlobalRID, DefinitionMsg]
    defs: tuple[DefinitionMsg, ...]
    added: set[GlobalRID]

    def __init__(self, frame: Frame, defs: tuple[DefinitionMsg, ...]):
        self.incoming = frame.incoming_defs
        self.defs = defs

    def __enter__(self):
        incoming = self.incoming
        self.added = added = set()
        for msg in self.defs:
            grid = msg.rid
            if grid not in incoming:
                incoming[grid] = msg
                added.add(grid)

    def __exit__(self, exc_type, exc_val, exc_tb):
        incoming = self.incoming
        for grid in self.added:
            incoming.pop(grid)

################################################################################

const_fn = CONSTANT_ENCODERS.get
atom_fn = SIMPLE_ENCODERS.get
compound_fn = COMPLEX_ENCODERS.get

################################################################################

def enc_list_iter(obj, enc):
    return enc.enc_any(list(obj))

def enc_tuple_iter(obj, enc):
    return enc.enc_any(tuple(obj))

def enc_set_iter(obj, enc):
    return enc.enc_any(set(obj))

################################################################################

def fmt_obj_stack(objs) -> str:
    lines = []
    add = lines.append
    for obj in objs:
        f_obj = '<error>'
        try:
            f_obj = f_object_id(obj)
        except:
            pass
        add('  during ' + f_obj)
    return '\n'.join(lines)

def fmt_id_set(ids: set[int]) -> str:
    text = ' '.join(f'0x{i:x}' for i in list(ids)[:16])
    return text if len(ids) < 16 else text + ' ...'

################################################################################

COMPLEX_ENCODERS[ListIterType] = enc_list_iter
COMPLEX_ENCODERS[TupleIterType] = enc_tuple_iter
COMPLEX_ENCODERS[SetIterType] = enc_set_iter
COMPLEX_ENCODERS[DictKeysType] = enc_list_iter
COMPLEX_ENCODERS[DictValuesType] = enc_list_iter
COMPLEX_ENCODERS[DictItemsType] = enc_list_iter

# uncomment to see mappings from literal classes to their compound encounders
# for k, v in COMPLEX_ENCODERS.items():
#     print(k.__name__.ljust(50), v)

################################################################################

frame_logger = logging.getLogger(__file__)

ERROR_BAR = P.RED('█' * 19)



# fmt: off

from typing import Protocol, NoReturn, Iterable as Iter, TYPE_CHECKING

from ..alias import *
from ..kinds import *

__all__ = [
    'EncoderP',
    'EncoderContextP',
    'DecoderP',
    'DecoderContextP',
    'CodecP',
    'NULL_CODEC'
]

################################################################################

if TYPE_CHECKING:
    from ..repl import ReplP
    from .term import TermMsg, TermPassByValMsg
    from .term_resource import ResourceMsg, LocalResourceMsg, RemoteResourceMsg, SystemResourceMsg
    from .resource_def import DefinitionMsg
    from .term_exception import ExceptionMsg
    from .msg_aliases import ArgsMsg, KwargsMsg, MethodsMsg, AnnotationsMsg

################################################################################

class EncoderP(Protocol):
    """
    Protocol (implemented by `warp.Frame`) that covers encoding of python values
    into messages.

    `enc_loc` only encodes resources that are local (NOT virtual).
    `enc_rem` only encodes resources that are remote (ARE virtual).
    `enc_sys` only encodes resources that are builtin system resources.

    `enc_args`, `enc_kwargs` drop ARG_DEFAULT values.
    """

    def enc_any(self, term: TermT, /)                -> 'TermMsg': ...

    def enc_exception(self, exc: BaseException, /)   -> 'ExceptionMsg': ...
    def enc_value(self, val: ValueT, /)              -> 'TermPassByValMsg': ...

    def enc_object(self, obj: ObjectT, /)            -> 'ResourceMsg': ...
    def enc_class(self, cls: ClassT, /)              -> 'ResourceMsg': ...
    def enc_type(self, cls: TypeT, /)                -> 'ResourceMsg': ...
    def enc_function(self, fun: FunctionT, /)        -> 'ResourceMsg': ...
    def enc_module(self, mod: ModuleT, /)            -> 'ResourceMsg': ...
    def enc_future(self, fut: FutureT, /)            -> 'ResourceMsg': ...

    def enc_resource(self, res: ResourceT, /)        -> 'ResourceMsg': ...
    def enc_local_resource(self, res: ResourceT, /)  -> 'LocalResourceMsg': ...
    def enc_remote_resource(self, res: ResourceT, /) -> 'RemoteResourceMsg': ...
    def enc_system_resource(self, res: ResourceT, /) -> 'SystemResourceMsg': ...

    def enc_sequence(self, seq: Iter[TermT], /)      -> 'Tup[TermMsg]': ...
    def enc_record(self, dct: Rec[TermT], /)         -> 'Rec[TermMsg]': ...

    def enc_args(self, tup: ArgsT, /)                -> 'ArgsMsg': ...
    def enc_kwargs(self, rec: KwargsT, /)            -> 'KwargsMsg': ...

    def enc_methods(self, rec: MethodsT, /)          -> 'MethodsMsg': ...
    def enc_annotations(self, rec: AnnotationsT, /)  -> 'AnnotationsMsg': ...

    def enc_owner(self)                              -> 'ResourceMsg': ...

    def enc_context(self) -> 'EncoderContextP': ...

    # this takes an existing future and returns a future_id for it, used when describing futures
    # as FutureData objects
    def future_to_id(self, future: FutureT, /) -> FutureID: ...


################################################################################

class EncoderContextP(Protocol):
    """
    A temporary context in which new definitions that are entrained during
    encoding can be found at the end of encoding.
    """

    def __enter__(self) -> None: ...
    def __exit__(self, exc_type, exc_val, exc_tb) -> None: ...

    def enc_context_defs(self) -> 'Tup[DefinitionMsg]': ...


################################################################################

class DecoderP(Protocol):
    """
    Protocol (implemented by `warp.Frame`) that covers decoding of python values
    into messages.
    """

    def dec_any(self, msg: 'TermMsg', /)                       -> TermT: ...

    def dec_value(self, msg: 'TermPassByValMsg', /)            -> ValueT: ...
    def dec_exception(self, msg: 'ExceptionMsg', /)            -> BaseException: ...

    def dec_object(self, msg: 'ResourceMsg', /)                -> ObjectT: ...
    def dec_class(self, msg: 'ResourceMsg', /)                 -> ClassT: ...
    def dec_type(self, msg: 'TermMsg', /)                      -> TypeT: ...
    def dec_function(self, msg: 'ResourceMsg', /)              -> FunctionT: ...
    def dec_module(self, msg: 'ResourceMsg', /)                -> ModuleT: ...
    def dec_future(self, rec: 'LocalResourceMsg', /)           -> FutureT: ...

    def dec_resource(self, msg: 'ResourceMsg', /)              -> ResourceT: ...
    def dec_local_resource(self, msg: 'LocalResourceMsg', /)   -> ResourceT: ...
    def dec_remote_resource(self, msg: 'RemoteResourceMsg', /) -> ResourceT: ...
    def dec_system_resource(self, msg: 'SystemResourceMsg', /) -> ResourceT: ...

    def dec_sequence(self, seq: 'Iter[TermMsg]', /)            -> Tup[TermT]: ...
    def dec_record(self, rec: 'Rec[TermMsg]', /)               -> Rec[TermT]: ...

    def dec_args(self, tup: 'ArgsMsg', /)                      -> ArgsT: ...
    def dec_kwargs(self, rec: 'KwargsMsg', /)                  -> KwargsT: ...
    def dec_methods(self, rec: 'MethodsMsg', /)                -> MethodsT: ...
    def dec_annotations(self, rec: 'AnnotationsMsg', /)        -> AnnotationsT: ...

    def get_repl(self) -> 'ReplP | None': ...

    def dec_context(self, msgs: 'Tup[DefinitionMsg]', /) -> 'DecoderContextP': ...

    # this creates or gets a future with a given future_id, used when taking
    # FutureData objects and turning them into futures. this will either be a
    # new totally virtual future, or an existing future that has a matching ID
    def future_from_id(self, future_id: FutureID, /) -> FutureT: ...


################################################################################

class DecoderContextP(Protocol):
    """
    A temporary context in which messages are decoded in the context of
    auxiliary definitions, which do not have to be in any particular order.
    """

    def __enter__(self) -> None: ...
    def __exit__(self, exc_type, exc_val, exc_tb) -> None: ...


################################################################################

class CodecP(EncoderP, DecoderP):
    """
    Combines encoding and decoding into one Protocol.
    """
    ...


################################################################################

def _none(*args) -> None:
    return None

def _raise(*args) -> NoReturn:
    raise RuntimeError("no encoder available")

class NullCodec(CodecP):

    dec_any              = _none
    dec_value            = _raise
    dec_exception        = _raise
    dec_sequence         = _raise
    dec_record           = _raise
    dec_object           = _raise
    dec_class            = _raise
    dec_type             = _raise
    dec_function         = _raise
    dec_module           = _raise
    dec_future           = _raise
    dec_resource         = _raise
    dec_local_resource   = _raise
    dec_remote_resource  = _raise
    dec_system_resource  = _raise
    dec_args             = _raise
    dec_kwargs           = _raise
    dec_methods          = _raise
    dec_annotations      = _raise

    enc_any              = _raise
    enc_exception        = _raise
    enc_value            = _raise
    enc_object           = _raise
    enc_class            = _raise
    enc_type             = _raise
    enc_function         = _raise
    enc_module           = _raise
    enc_future           = _raise
    enc_resource         = _raise
    enc_local_resource   = _raise
    enc_remote_resource  = _raise
    enc_system_resource  = _raise
    enc_sequence         = _raise
    enc_record           = _raise
    enc_args             = _raise
    enc_kwargs           = _raise
    enc_methods          = _raise
    enc_annotations      = _raise

    enc_owner            = _none
    get_repl             = _none

    future_from_id       = _raise
    future_to_id         = _raise

    def enc_context(self) -> 'EncoderContextP':
        return self

    def dec_context(self, msgs: 'Tup[DefinitionMsg]', /):
        return self

    def enc_context_defs(self) -> 'Tup[DefinitionMsg]':
        return ()

    def __enter__(self) -> None: ...
    def __exit__(self, exc_type, exc_val, exc_tb) -> None: ...


NULL_CODEC = NullCodec()



# EXCLUDE_FROM_SDK_BUILD
# fmt: off
# ruff: noqa

# NOTE: this is to ensure that componentize will acquire these modules

import agentica_internal.core.color
import agentica_internal.core.debug
import agentica_internal.core.fmt
import agentica_internal.core.fns
import agentica_internal.core.futures
import agentica_internal.core.hashing
import agentica_internal.core.json
import agentica_internal.core.log
import agentica_internal.core.make
import agentica_internal.core.print
import agentica_internal.core.queues
import agentica_internal.core.recursion
import agentica_internal.core.result
import agentica_internal.core.sentinels
import agentica_internal.core.strs
import agentica_internal.core.type
import agentica_internal.core.utils

import agentica_internal.cpython.alias
import agentica_internal.cpython.classes.anno
import agentica_internal.cpython.classes.sys
import agentica_internal.cpython.code
import agentica_internal.cpython.iters
import agentica_internal.cpython.frame
import agentica_internal.cpython.function
import agentica_internal.cpython.ids
import agentica_internal.cpython.inspect
import agentica_internal.cpython.module
import agentica_internal.cpython.slots

# ensure the shed modules are known about, they will be
# cleared from sys.modules when the shed actually loads them though!
import agentica_internal.cpython.shed
import agentica_internal.cpython.shed.load
import agentica_internal.cpython.shed._math
import agentica_internal.cpython.shed._builtins

import agentica_internal.warpc.request.all
import agentica_internal.warpc.resource.all
import agentica_internal.warpc.resource.virtual_builtin

import msgspec
import rich

__all__ = []



from agentica_internal.core.mixin import mixin

from . import b
from .b import *

__all__ = 'foo', 'bar', 'decorated_caller', 'helper', 'get_registry', 'is_feature_c'

mixin(b)


def helper() -> str:
    return 'c.helper'


def foo() -> str:
    return 'c.foo'


def is_feature_c() -> bool:
    return True



# fmt: off

from collections.abc import Iterable
from typing import Self

__all__ = [
    'AnyChain',
    'AnyTree',
    'AnyNat',
    'Chain',
    'ChainLink',
    'ChainEnd',
    'Tree',
    'TreeNode',
    'TreeLeaf',
    'Nat',
    'NatOdd',
    'NatEven',
    'NatZero',
    'RECURSIVE_CLASSES',
    'RECURSIVE_ALIASES',
]


class Chain:
    """Chain docstring"""

    def step(self) -> Self: ...


class ChainLink(Chain):
    """ChainLink docstring"""

    n: Chain

    def __init__(self, n: Chain):
        self.n = n

    def step(self) -> Chain:
        return self.n


class ChainEnd(Chain):
    """ChainEnd docstring"""

    def __init__(self): ...

    def step(self) -> Self:
        return self


class Tree:
    """Tree docstring"""

    def leaves(self) -> Iterable['TreeLeaf']: ...


class TreeNode(Tree):
    """TreeNode docstring"""

    l: Tree
    r: Tree

    def __init__(self, l: Tree, r: Tree):
        self.l = l
        self.r = r

    def leaves(self) -> Iterable['TreeLeaf']:
        yield from self.l.leaves()
        yield from self.r.leaves()


class TreeLeaf(Tree):
    """TreeLeaf docstring"""

    def __init__(self): ...

    def leaves(self) -> Iterable['TreeLeaf']:
        yield self


class Nat:
    """Nat docstring"""
    i: int

    def __init__(self, i: int):
        self.i = i

    def dec(self) -> Self: ...
    def inc(self) -> Self: ...


class NatOdd(Nat):
    """NatOdd docstring"""

    def dec(self) -> 'NatEven':
        return NatEven(self.i - 1) if self.i > 1 else NatZero(0)

    def inc(self) -> 'NatEven':
        return NatEven(self.i + 1)


class NatEven(Nat):
    """NatEven docstring"""

    def dec(self) -> 'NatOdd | NatZero':
        return NatOdd(self.i - 1)

    def inc(self) -> NatOdd:
        return NatOdd(self.i + 1)


class NatZero(Nat):
    """NatZero docstring"""

    def __init__(self):
        super().__init__(0)

    def dec(self) -> Self: ...
    def inc(self) -> NatEven: ...


type AnyChain = ChainLink | ChainEnd
type AnyTree = TreeNode | TreeLeaf
type AnyNat = NatOdd | NatEven | NatZero


RECURSIVE_CLASSES: list[type] = [
    Chain,
    ChainLink,
    ChainEnd,
    Tree,
    TreeNode,
    TreeLeaf,
    Nat,
    NatOdd,
    NatEven,
    NatZero,
]

RECURSIVE_ALIASES: list = [AnyChain, AnyTree, AnyNat]

RECURSIVE_INSTANCES: list = [
    ChainEnd(),
    ChainLink(ChainEnd()),
    ChainLink(ChainLink(ChainEnd())),
    TreeLeaf(),
    TreeNode(TreeLeaf(), TreeLeaf()),
    TreeNode(TreeNode(TreeLeaf(), TreeLeaf()), TreeLeaf()),
    NatZero(),
    NatOdd(1),
    NatEven(2),
]



from agentica_internal.cpython.ids import (
    SYS_CLASS_ID_DICT,
    SYS_FUNCTION_ID_DICT,
    SYS_OBJECT_ID_DICT,
)
from agentica_internal.warpc.system import (
    CLS_OFFSET,
    FUN_OFFSET,
    OBJ_OFFSET,
    SRID_TO_CLS,
    SRID_TO_FUN,
    SRID_TO_OBJ,
)


def verify_system_class_ids():
    for cls_name, cls_id in SYS_CLASS_ID_DICT.items():
        srid = cls_id + CLS_OFFSET
        assert srid in SRID_TO_CLS, f"system class {cls_name} not registered"


def verify_system_function_ids():
    for fun_name, fun_id in SYS_FUNCTION_ID_DICT.items():
        srid = fun_id + FUN_OFFSET
        assert srid in SRID_TO_FUN, f"system function {fun_name} not registered"


RENAMED = {'IN__is_coroutine_mark'}


def verify_system_object_ids():
    for obj_name, obj_id in SYS_OBJECT_ID_DICT.items():
        srid = obj_id + OBJ_OFFSET
        if obj_name in RENAMED:
            continue
        assert srid in SRID_TO_OBJ, f"system object {obj_name} not registered"


def verify_system_tables_disjoint():
    funs = set(map(id, SRID_TO_FUN.values()))
    clss = set(map(id, SRID_TO_CLS.values()))
    objs = set(map(id, SRID_TO_OBJ.values()))
    assert funs.isdisjoint(clss), f"classes and funtions not disjoint"
    assert objs.isdisjoint(clss), f"classes and objects not disjoint"
    assert funs.isdisjoint(objs), f"functions and objects not disjoint"


if __name__ == '__main__':
    verify_system_class_ids()
    verify_system_function_ids()
    verify_system_object_ids()
    verify_system_tables_disjoint()



# fmt: off
"""
This a hand-curated typeshed file for the builtin `typing` module, specifically the exposed classes and type objections
and some of their non-exported base classes.
"""

import collections.abc
import sys
from types import CodeType as CODETYPE
from types import NoneType as NONETYPE
from typing import Any as UNTYPED
from typing import Never as NEVER
from typing import NoReturn as NORETURN

####################################################################################################

__all__ = [
    # namespaces
    'META', 'PRIVATE',

    # actually exposed by `typing` module
    'get_origin', 'get_args', 'get_type_hints',
    'TypeAliasType', 'TypeVar', 'TypeVarTuple', 'ParamSpec', 'ParamSpecArgs', 'ParamSpecKwargs', 'ForwardRef',
    'Any', 'NoDefault', 'NewType', 'Protocol', 'Annotated', 'TypedDict',
    'NoReturn', 'Never', 'Self', 'LiteralString', 'ClassVar', 'Final', 'Union', 'Optional', 'Unpack',
    'Literal', 'TypeAlias', 'Concatenate', 'TypeGuard', 'TypeIs', 'Required', 'NotRequired', 'ReadOnly',

    # deprecated aliases to collections.abc.XXX
    'Hashable', 'Awaitable', 'Coroutine', 'AsyncIterable', 'AsyncIterator', 'Iterable', 'Iterator', 'Reversible',
    'Sized', 'Container', 'Collection', 'Callable', 'AbstractSet', 'MutableSet', 'Mapping', 'MutableMapping',
    'Sequence', 'MutableSequence', 'List', 'Deque', 'Set', 'FrozenSet', 'MappingView', 'KeysView', 'ItemsView',
    'ValuesView', 'Dict', 'DefaultDict', 'OrderedDict', 'Counter', 'ChainMap', 'Generator', 'AsyncGenerator', 'Type',
]

####################################################################################################

PY13 = sys.version_info >= (3, 13)

type GenericTypeObj = CPythonGenericAlias | CPythonUnionType | _BaseGenericAlias
type SymbolicTypeObj = TypeAliasType | TypeVar | ParamSpec | _SpecialForm | _AnyMeta
type TypingObj = GenericTypeObj | SymbolicTypeObj

type TypeObj = type | TypingObj
type TypeObjs = tuple[TypeObj, ...]

type TypeParam = TypeVar | ParamSpec | TypeVarTuple
type TypeParams = tuple[TypeParam, ...]

####################################################################################################


def get_origin(obj: UNTYPED) -> TypeObj | None:
    return None


def get_args(obj: UNTYPED) -> TypeObjs:
    return ()


def get_type_hints(obj: UNTYPED) -> dict[str, TypeObj]:
    return {}


####################################################################################################


def _idfunc[X](arg: X, *args, **kwargs) -> X:
    return arg


####################################################################################################


class NoDefaultType:
    """Indicates a TypeParam has no default."""


NoDefault = NoDefaultType()


class _Sentinel:
    """Internal sentinel singleton Class."""


_sentinel: _Sentinel = _Sentinel()


class _TypingEllipsis:
    """Internal placeholder for ... (ellipsis)."""


class _AnyMeta(type):
    def __instancecheck__(self, obj) -> bool: ...
    def __repr__(self) -> str: ...


class Any(metaclass=_AnyMeta):
    def __new__(cls, *args, **kwargs): ...


####################################################################################################


class _SpecialForm:
    _name: str
    __name__: str

    def _getitem(self): ...
    def __getattr__(self, item): ...
    def __mro_entries__(self, bases) -> NORETURN: ...
    def __repr__(self) -> str: ...

    def __reduce__(self) -> str | tuple[Any, ...]: ...
    def __call__(self, *args, **kwds) -> NORETURN: ...

    def __or__(self, other): ...
    def __ror__(self, other): ...
    def __instancecheck__(self, obj) -> NORETURN: ...
    def __subclasscheck__(self, cls): ...

    # explicitly tell python item type is unknown
    def __getitem__(self, parameters) -> TypingObj: ...


###############################################################################


class CPythonUnionType:
    __args__: TypeObjs

    def __or__(self, value: Any, /) -> 'CPythonUnionType': ...
    def __ror__(self, value: Any, /) -> 'CPythonUnionType': ...
    def __eq__(self, value: object, /) -> bool: ...
    def __hash__(self) -> int: ...


####################################################################################################


class CPythonGenericAlias:
    """
    types.GenericAlias is defined in cypthon/Objects/generalicaliasobject.c and corresponds to
    generic aliases for builtins, like list[int], dict[str, bool], etc. and has instance fields:

    __origin__:     TypeObj
    __args__:       Tup[TypeObj]
    __parameters__: Tup[TypeVar | ParamSpec | TypeVarTuple]
    __unpacked__:   ???
    """

    __origin__: TypeObj
    __args__: TypeObjs
    __unpacked__: UNTYPED
    __parameters__: TypeParams

    @property
    def __typing_unpacked_tuple_args__(self):
        return None

    def __mro_entries__(self, bases):
        return ()

    def __copy__(self) -> 'CPythonGenericAlias':
        return self

    def __deepcopy__(self) -> 'CPythonGenericAlias':
        return self


####################################################################################################


class TypeAliasType:
    """
    typing.TypeAliasType is the type of the values declared via `type alias = int`, and has instance fields:

    __value__:       TypeObj
    __type_params__: Tup[TypeVar | ParamSpec | TypeVarTuple]
    __parameters__:  Tup[TypeObj]
    __name__:        str
    __module__:      str | None
    """

    def __init__(self, name: str, value: UNTYPED, *, type_params: TypeParams = ()): ...

    @property
    def __value__(self) -> TypeObj:
        return object

    @property
    def __type_params__(self) -> TypeParams:
        return ()

    @property
    def __parameters__(self) -> TypeObjs:
        return ()

    @property
    def __name__(self) -> str:
        return 'name'

    __module__: str = 'module'


####################################################################################################


class TypeVar:
    """
    typing.TypeVar is the type of values created when you write e.g. `class Foo[X]: ...`, and has instance fields:

    __bound__:         TypeObj | None
    __constraints__:   Tup[TypeObj]
    __contravariant__: bool
    __covariant__:     bool
    __default__:       TypeObj | NoDefaultType
    __name__:          str
    """

    __bound__: TypeObj | None
    __constraints__: TypeObjs
    __contravariant__: bool
    __covariant__: bool
    __default__: TypeObj | NoDefaultType

    @property
    def __name__(self) -> str:
        return 'name'


####################################################################################################


class TypeVarTuple:
    """
    typing.TypeVarTuple is the type of values created when you write e.g. `class Foo[*X]: ...`, and has instance fields:
    __default__:       TypeObj | NoDefaultType
    __name__:          str
    has_default:       bool
    """

    @property
    def __name__(self) -> str:
        return 'name'

    @property
    def __default__(self) -> TypeObj | NoDefaultType:
        return NoDefault

    def has_default(self) -> bool:
        return False

    def __init__(self, name: str, *, default: UNTYPED = ...): ...
    def __iter__(self) -> UNTYPED:
        return ()

    def __typing_subst__(self, arg: NEVER) -> NEVER: ...
    def __typing_prepare_subst__(self, alias: UNTYPED, args: UNTYPED) -> tuple[UNTYPED, ...]: ...


####################################################################################################


class ParamSpecArgs:
    @property
    def __origin__(self) -> 'ParamSpec': ...

    def __init__(self, origin: 'ParamSpec'): ...
    def __eq__(self, other: object) -> bool: ...


class ParamSpecKwargs:
    @property
    def __origin__(self) -> 'ParamSpec': ...

    def __init__(self, origin: 'ParamSpec'): ...
    def __eq__(self, other: object) -> bool: ...


class ParamSpec:
    @property
    def __name__(self) -> str:
        return 'name'

    @property
    def __bound__(self) -> UNTYPED:
        return None

    @property
    def __covariant__(self) -> bool:
        return True

    @property
    def __contravariant__(self) -> bool:
        return False

    @property
    def __infer_variance__(self) -> bool:
        return True

    @property
    def __default__(self) -> TypeObj | NoDefaultType:
        return NoDefault

    def has_default(self) -> bool:
        return False

    def __init__(
        self,
        name: str,
        *,
        bound: UNTYPED = None,
        contravariant: bool = False,
        covariant: bool = False,
        infer_variance: bool = False,
        default: UNTYPED = ...,
    ):
        pass

    @staticmethod
    def __new__(*args, **kwargs): ...
    @property
    def args(self) -> ParamSpecArgs: ...

    @property
    def kwargs(self) -> ParamSpecKwargs: ...

    def __typing_subst__(self, arg: UNTYPED) -> UNTYPED: ...
    def __typing_prepare_subst__(self, alias: UNTYPED, args: UNTYPED) -> tuple[UNTYPED, ...]: ...

    def __or__(self, right: UNTYPED) -> _SpecialForm: ...
    def __ror__(self, left: UNTYPED) -> _SpecialForm: ...


####################################################################################################


class ForwardRef:
    slots = (
        '__forward_arg__',
        '__forward_code__',
        '__forward_evaluated__',
        '__forward_value__',
        '__forward_is_argument__',
        '__forward_is_class__',
        '__forward_module__',
    )

    __forward_arg__: str
    __forward_code__: CODETYPE
    __forward_evaluated__: bool
    __forward_value__: UNTYPED | None
    __forward_is_argument__: bool
    __forward_is_class__: bool
    __forward_module__: UNTYPED | None

    def __init__(
        self,
        arg: str,
        is_argument: bool = True,
        module: Any | None = None,
        *,
        is_class: bool = False,
    ): ...

    def _evaluate(
        self,
        globalns: dict[str, UNTYPED] | None,
        localns: dict[str, UNTYPED] | None,
        type_params: TypeParams,
        *,
        recursive_guard: frozenset[str],
    ) -> UNTYPED: ...

    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __or__(self, other: object) -> _SpecialForm: ...
    def __ror__(self, other: object) -> _SpecialForm: ...
    def __repr__(self) -> str: ...


####################################################################################################


class _BaseGenericAlias:
    """The central part of the internal API.

    This represents a generic version of type 'origin' with type arguments 'params'.
    There are two kind of these aliases: user defined and special. The special ones
    are wrappers around builtin collections and ABCs in collections.abc. These must
    have 'name' always set. If 'inst' is False, then the alias can't be instantiated;
    this is used by e.g. typing.List and typing.Dict.
    """

    _inst: bool  # whether it can be instantiated
    _name: str  # name of the generic, proxied by .__name__
    __origin__: type  # underlying class

    def __init__(self, origin, *, inst=True, name=None): ...
    def __call__(self, *args, **kwargs) -> type: ...
    def __mro_entries__(self, bases) -> tuple[type, ...]: ...
    def __getattr__(self, attr): ...
    def __setattr__(self, attr, val): ...
    def __instancecheck__(self, obj) -> bool: ...
    def __subclasscheck__(self, cls): ...
    def __dir__(self) -> list[str]: ...


# Special typing constructs Union, Optional, Generic, Callable and Tuple
# use three special attributes for internal bookkeeping of generic types:
# * __parameters__ is a tuple of unique free type parameters of a generic
#   type, for example, Dict[T, T].__parameters__ == (T,);
# * __origin__ keeps a reference to a type that was subscripted,
#   e.g., Union[T, int].__origin__ == Union, or the non-generic version of
#   the type.
# * __args__ is a tuple of all arguments used in subscripting,
#   e.g., Dict[T, int].__args__ == (T, int).


####################################################################################################


class _GenericAlias(_BaseGenericAlias):
    """The type of parameterized generics.

    That is, for example, `type(List[int])` is `GenericAlias`.

    Objects which are instances of this class include:
    * Parameterized container types, e.g. `Tuple[int]`, `List[int]`.
     * Note that native container types, e.g. `tuple`, `list`, use
       `types.GenericAlias` instead.
    * Parameterized classes:
        class C[T]: pass
        # C[int] is a GenericAlias
    * `Callable` aliases, generic `Callable` aliases, and
      parameterized `Callable` aliases:
        T = TypeVar('T')
        # _CallableGenericAlias inherits from GenericAlias.
        A = Callable[[], None]  # _CallableGenericAlias
        B = Callable[[T], None]  # _CallableGenericAlias
        C = B[int]  # _CallableGenericAlias
    * Parameterized `Final`, `ClassVar`, `TypeGuard`, and `TypeIs`:
        # All GenericAlias
        Final[int]
        ClassVar[float]
        TypeGuard[bool]
        TypeIs[range]
    """

    __parameters__: TypeObjs
    __value__: TypeObj

    def __eq__(self, other) -> bool: ...
    def __hash__(self) -> int: ...
    def __or__(self, right): ...
    def __getitem__(self, args): ...
    def _determine_new_args(self, args): ...
    def _make_substitution(self, args, new_arg_by_param): ...
    def copy_with(self, params) -> '_GenericAlias': ...
    def __repr__(self) -> str: ...
    def __reduce__(self) -> str | tuple[Any, ...]: ...
    def __mro_entries__(self, bases) -> tuple[type, ...]: ...
    def __iter__(self): ...


####################################################################################################


class _SpecialGenericAlias(_BaseGenericAlias):
    _nparams: int
    _defaults: tuple

    def __init__(self, origin, nparams: int, *, inst=True, name=None, defaults=()):
        super().__init__(origin)
        pass

    def __getitem__(self, params): ...
    def copy_with(self, params) -> '_SpecialGenericAlias': ...
    def __repr__(self) -> str: ...
    def __subclasscheck__(self, cls): ...
    def __reduce__(self) -> str | tuple[Any, ...]: ...
    def __or__(self, right): ...
    def __ror__(self, left): ...


####################################################################################################


class _CallableGenericAlias(_GenericAlias):
    def __repr__(self) -> str: ...
    def __reduce__(self) -> str | tuple[Any, ...]: ...


####################################################################################################


if PY13:
    # this proxies the class created by collections.abc.Callable[XXX, YYY]
    class _AbcCallableGenericAlias(CPythonGenericAlias):
        __slots__ = ()

        def __new__(cls, origin: UNTYPED, args: tuple[UNTYPED, UNTYPED]): ...

        def __repr__(self) -> str: ...

        def __getitem__(self, item) -> '_AbcCallableGenericAlias': ...


####################################################################################################


class _CallableType(_SpecialGenericAlias):
    def copy_with(self, params) -> '_CallableType': ...
    def __getitem__(self, params): ...
    def __getitem_inner__(self, params): ...


####################################################################################################


class _TupleType(_SpecialGenericAlias):
    def __getitem__(self, params): ...


####################################################################################################


class _UnionGenericAlias(_GenericAlias):
    def copy_with(self, params) -> '_UnionGenericAlias': ...
    def __eq__(self, other) -> bool: ...
    def __hash__(self) -> int: ...
    def __repr__(self) -> str: ...
    def __instancecheck__(self, obj) -> bool: ...
    def __subclasscheck__(self, cls): ...
    def __reduce__(self) -> str | tuple[Any, ...]: ...


####################################################################################################


class _LiteralGenericAlias(_GenericAlias):
    def __eq__(self, other) -> bool: ...
    def __hash__(self) -> int: ...


####################################################################################################


class _ConcatenateGenericAlias(_GenericAlias):
    def copy_with(self, params) -> _GenericAlias: ...


####################################################################################################


class _UnpackGenericAlias(_GenericAlias):
    def __repr__(self) -> str: ...

    def __getitem__(self, args): ...

    @property
    def __typing_unpacked_tuple_args__(self) -> UNTYPED:
        return None

    @property
    def __typing_is_unpacked_typevartuple__(self) -> UNTYPED:
        return None


####################################################################################################


class _AnnotatedAlias(_GenericAlias):
    """Runtime representation of an annotated type.

    At its core 'Annotated[t, dec1, dec2, ...]' is an alias for the type 't'
    with extra annotations. The alias behaves like a normal typing alias.
    Instantiating is the same as instantiating the underlying type; binding
    it to types is also the same.

    The metadata itself is stored in a '__metadata__' attribute as a tuple.
    """

    def __init__(self, origin, metadata):
        super().__init__(origin)

    def copy_with(self, params) -> '_AnnotatedAlias': ...
    def __repr__(self) -> str: ...
    def __reduce__(self) -> str | tuple[Any, ...]: ...
    def __eq__(self, other) -> bool: ...
    def __hash__(self) -> int: ...
    def __getattr__(self, attr): ...
    def __mro_entries__(self, bases) -> tuple[type, ...]: ...


class _AnnotatedSpecialForm(_SpecialForm):
    def __getitem__(self, parameters) -> _AnnotatedAlias: ...


Annotated = _AnnotatedSpecialForm()

####################################################################################################


class _TypedDictMeta(type):
    def __new__(cls, name, bases, ns, total=True): ...


def TypedDict(typename, fields=_sentinel, /, *, total=True) -> _TypedDictMeta: ...


TypedDictType: _TypedDictMeta


####################################################################################################


class NewType:
    """NewType creates simple unique types with almost zero runtime overhead."""

    __call__ = _idfunc

    def __init__(self, name, tp): ...
    def __mro_entries__(self, bases) -> tuple[type, ...]: ...
    def __repr__(self) -> str: ...
    def __reduce__(self) -> str | tuple[Any, ...]: ...
    def __or__(self, other): ...
    def __ror__(self, other): ...


####################################################################################################


class Protocol:
    """Base class for protocol classes.

    Protocol classes are defined as::

        class Proto(Protocol):
            def meth(self) -> int:
                ...

    Such classes are primarily used with static type checkers that recognize
    structural subtyping (static duck-typing).

    For example::

        class C:
            def meth(self) -> int:
                return 0

        def func(x: Proto) -> int:
            return x.meth()

        func(C())  # Passes static type check

    See PEP 544 for details. Protocol classes decorated with
    @typing.runtime_checkable act as simple-minded runtime protocols that check
    only the presence of given attributes, ignoring their type signatures.
    Protocol classes can be generic, they are defined as::

        class GenProto[T](Protocol):
            def meth(self) -> T:
                ...
    """

    __slots__ = ()
    _is_protocol = True
    _is_runtime_protocol = False

    def __init_subclass__(cls, *args, **kwargs): ...


####################################################################################################


def _eval_type(
    t: UNTYPED, globalns, localns, type_params=_sentinel, *, recursive_guard=frozenset()
) -> TypeObj: ...
def _type_repr(arg: UNTYPED) -> str:
    return 'type'


####################################################################################################

Hashable = _SpecialGenericAlias(collections.abc.Hashable, 0)  # Not generic.
Awaitable = _SpecialGenericAlias(collections.abc.Awaitable, 1)
Coroutine = _SpecialGenericAlias(collections.abc.Coroutine, 3)
AsyncIterable = _SpecialGenericAlias(collections.abc.AsyncIterable, 1)
AsyncIterator = _SpecialGenericAlias(collections.abc.AsyncIterator, 1)
Iterable = _SpecialGenericAlias(collections.abc.Iterable, 1)
Iterator = _SpecialGenericAlias(collections.abc.Iterator, 1)
Reversible = _SpecialGenericAlias(collections.abc.Reversible, 1)
Sized = _SpecialGenericAlias(collections.abc.Sized, 0)  # Not generic.
Container = _SpecialGenericAlias(collections.abc.Container, 1)
Collection = _SpecialGenericAlias(collections.abc.Collection, 1)
Callable = _CallableType(collections.abc.Callable, 2)
AbstractSet = _SpecialGenericAlias(collections.abc.Set, 1, name='AbstractSet')
MutableSet = _SpecialGenericAlias(collections.abc.MutableSet, 1)
Mapping = _SpecialGenericAlias(collections.abc.Mapping, 2)
MutableMapping = _SpecialGenericAlias(collections.abc.MutableMapping, 2)
Sequence = _SpecialGenericAlias(collections.abc.Sequence, 1)
MutableSequence = _SpecialGenericAlias(collections.abc.MutableSequence, 1)
List = _SpecialGenericAlias(list, 1, inst=False, name='List')
Deque = _SpecialGenericAlias(collections.deque, 1, name='Deque')
Set = _SpecialGenericAlias(set, 1, inst=False, name='Set')
FrozenSet = _SpecialGenericAlias(frozenset, 1, inst=False, name='FrozenSet')
MappingView = _SpecialGenericAlias(collections.abc.MappingView, 1)
KeysView = _SpecialGenericAlias(collections.abc.KeysView, 1)
ItemsView = _SpecialGenericAlias(collections.abc.ItemsView, 2)
ValuesView = _SpecialGenericAlias(collections.abc.ValuesView, 1)
Dict = _SpecialGenericAlias(dict, 2, inst=False, name='Dict')
DefaultDict = _SpecialGenericAlias(collections.defaultdict, 2, name='DefaultDict')
OrderedDict = _SpecialGenericAlias(collections.OrderedDict, 2)
Counter = _SpecialGenericAlias(collections.Counter, 1)
ChainMap = _SpecialGenericAlias(collections.ChainMap, 2)
Generator = _SpecialGenericAlias(collections.abc.Generator, 3, defaults=(NONETYPE, NONETYPE))
AsyncGenerator = _SpecialGenericAlias(collections.abc.AsyncGenerator, 2, defaults=(NONETYPE,))
Type = _SpecialGenericAlias(type, 1, inst=False, name='Type')

####################################################################################################


class PRIVATE:
    idfunc = _idfunc
    eval_type = _eval_type
    type_repr = _type_repr
    TypingEllipsis = _TypingEllipsis
    Sentinel = _Sentinel
    sentinel = _sentinel
    AnyMeta = _AnyMeta
    AnnotatedAlias = _AnnotatedAlias
    TypedDictMeta = _TypedDictMeta
    SpecialForm = _SpecialForm
    CPythonGenericAlias = CPythonGenericAlias
    CPythonUnionType = CPythonUnionType
    AbcCallableGenericAlias = _AbcCallableGenericAlias
    BaseGenericAlias = _BaseGenericAlias
    GenericAlias = _GenericAlias
    SpecialGenericAlias = _SpecialGenericAlias
    CallableGenericAlias = _CallableGenericAlias
    CallableType = _CallableType
    TupleType = _TupleType
    UnionGenericAlias = _UnionGenericAlias
    LiteralGenericAlias = _LiteralGenericAlias
    ConcatenateGenericAlias = _ConcatenateGenericAlias
    UnpackGenericAlias = _UnpackGenericAlias


####################################################################################################

# these don't exist, but the shed pretends they do help type inference understand
# what happens in various cases.

type _Forms = tuple[_SpecialForm, ...]


class AnyForm(_SpecialForm):
    OBJS: _Forms = ()

    def __getitem__(self, parameters) -> NORETURN: ...


class SymbolForm(_SpecialForm):
    OBJS: _Forms = ()

    def __getitem__(self, parameters) -> NORETURN: ...


class UnaryForm[T: TypingObj](_SpecialForm):
    OBJS: _Forms = ()

    def __getitem__(self, parameters) -> T: ...


class NaryForm[T: TypingObj](_SpecialForm):
    OBJS: _Forms = ()

    def __getitem__(self, parameters: tuple) -> T: ...


class WrapperForm(_SpecialForm):
    OBJS: _Forms = ()

    def __getitem__(self, parameters) -> _GenericAlias: ...


class GuardForm(WrapperForm):
    OBJS: _Forms = ()


class FieldForm(WrapperForm):
    OBJS: _Forms = ()


class ClsFieldForm(FieldForm):
    pass


class DctFieldForm(FieldForm):
    pass


####################################################################################################

NoReturn = SymbolForm()
Never = SymbolForm()
Self = SymbolForm()
LiteralString = SymbolForm()
Union = NaryForm[_UnionGenericAlias]()
Optional = UnaryForm[_UnionGenericAlias]()
Unpack = UnaryForm[_UnpackGenericAlias]()
Literal = NaryForm[_LiteralGenericAlias]()
TypeAlias = SymbolForm()
Concatenate = NaryForm[_ConcatenateGenericAlias]()

ClassVar = ClsFieldForm()
Final = ClsFieldForm()
Required = DctFieldForm()
NotRequired = DctFieldForm()
ReadOnly = DctFieldForm()
TypeGuard = GuardForm()
TypeIs = GuardForm()

####################################################################################################


class META:
    GenericTypeObj = GenericTypeObj
    SymbolicTypeObj = SymbolicTypeObj
    TypingObj = TypingObj
    TypeObj = TypeObj
    TypeObjs = TypeObjs
    TypeParam = TypeParam
    TypeParams = TypeParams
    AnyForm = AnyForm
    SymbolForm = SymbolForm
    GuardForm = GuardForm
    FieldForm = FieldForm
    WrapperForm = WrapperForm
    UnaryForm = UnaryForm
    NaryForm = NaryForm

    GENERIC_TYPE_CLASSES: tuple[type, ...] = ()
    SYMBOLIC_TYPE_CLASSES: tuple[type, ...] = ()
    TYPING_OBJ_CLASSES: tuple[type, ...] = ()
    TYPE_OBJ_CLASSES: tuple[type, ...] = ()

    class FORMS:
        SYMBOL: _Forms = (
            NoReturn,
            Never,
            Self,
            LiteralString,
            TypeAlias,
        )
        UNARY: _Forms = (
            Unpack,
            Optional,
        )
        NARY: _Forms = (
            Union,
            Literal,
            Concatenate,
        )
        GUARD: _Forms = (
            TypeGuard,
            TypeIs,
        )
        CLS_FIELD: _Forms = (
            ClassVar,
            Final,
        )
        DCT_FIELD: _Forms = (
            Required,
            NotRequired,
            ReadOnly,
        )
        FIELD: _Forms = CLS_FIELD + DCT_FIELD
        WRAPPER: _Forms = GUARD + FIELD
        ALL: _Forms = SYMBOL + WRAPPER + UNARY + NARY

    AnyForm.OBJS = FORMS.ALL
    SymbolForm.OBJS = FORMS.SYMBOL
    UnaryForm.OBJS = FORMS.UNARY
    NaryForm.OBJS = FORMS.NARY
    WrapperForm.OBJS = FORMS.WRAPPER
    GuardForm.OBJS = FORMS.WRAPPER
    FieldForm.OBJS = FORMS.FIELD
    ClsFieldForm.OBJS = FORMS.CLS_FIELD
    DctFieldForm.OBJS = FORMS.DCT_FIELD



# fmt: off

import asyncio
from sys import version_info

__all__ = [
    'Queue',
    'QueueEmpty',
    'QueueShutDown',
]

###############################################################################

if version_info >= (3, 13):
    Queue = asyncio.Queue

    QueueEmpty = asyncio.QueueEmpty

    QueueShutDown = asyncio.QueueShutDown

else:
    QueueEmpty = asyncio.QueueEmpty

    class QueueShutDown(Exception): ...

    class Queue(asyncio.Queue):
        _shut: bool

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._shut = False

        def shutdown(self):
            self._shut = True

        async def put(self, val):
            if self._shut:
                raise QueueShutDown()
            await super().put(val)

        def put_nowait(self, val):
            if self._shut:
                raise QueueShutDown()
            super().put_nowait(val)

        async def get(self):
            if self._shut and self.empty():
                raise QueueShutDown()
            return await super().get()

        def get_nowait(self):
            if self._shut and self.empty():
                raise QueueShutDown()
            return super().get_nowait()



from re import compile as re_compile
from typing import ClassVar, Self

__all__ = [
    'AnsiCode',
    'strip_ansi',
    'ansi_len',
    'ansi_ljust',
    'RESET',
    'RESET_BG',
    'RESET_FG',
    'WARN',
    'ERROR',
    'WHITE',
    'RED',
    'YELLOW',
    'GREEN',
    'BLUE',
    'MAGENTA',
    'CYAN',
    'BOLD',
    'ITALIC',
    'UNDER',
    'DIM',
    'GRAY_BG',
    'ICON',
]


####################################################################################################


class AnsiCode(str):
    def __new__(cls, c: str):
        return super().__new__(cls, c)

    def __call__(self, s: str, w: int = 0) -> str:
        return apply_ansi(self, s, w)

    def __or__(self, s: Self) -> Self:
        return AnsiCode(f'{self};{s}')

    def wrap(self) -> str:
        return wrap_ansi(self)

    # fmt: off
    RESET:    ClassVar['AnsiCode']
    RESET_BG: ClassVar['AnsiCode']
    RESET_FG: ClassVar['AnsiCode']
    WARN:     ClassVar['AnsiCode']
    ERROR:    ClassVar['AnsiCode']
    WHITE:    ClassVar['AnsiCode']
    RED:      ClassVar['AnsiCode']
    YELLOW:   ClassVar['AnsiCode']
    GREEN:    ClassVar['AnsiCode']
    BLUE:     ClassVar['AnsiCode']
    MAGENTA:  ClassVar['AnsiCode']
    CYAN:     ClassVar['AnsiCode']
    BOLD:     ClassVar['AnsiCode']
    ITALIC:   ClassVar['AnsiCode']
    UNDER:    ClassVar['AnsiCode']
    DIM:      ClassVar['AnsiCode']
    GRAY_BG:  ClassVar['AnsiCode']
    ICON:     ClassVar['AnsiCode']
    # fmt: on


####################################################################################################


ansi_code_re = re_compile(r'\033\[[0-9;]+m')


def strip_ansi(text: str) -> str:
    return ansi_code_re.sub('', text)


def apply_ansi(code: AnsiCode, text: str, w: int = 0) -> str:
    return f'\033[{code}m{text.ljust(w)}\033[0m'


def wrap_ansi(code: AnsiCode) -> str:
    return f'\033[{code}m'


def ansi_len(s: str) -> int:
    if '\033[' not in s:
        return len(s)
    return len(strip_ansi(s))


def ansi_ljust(s: str, n: int) -> str:
    if '\033[' not in s:
        return s.ljust(n)
    m = len(strip_ansi(s))
    if m < n:
        return s + ' ' * (n - m)
    return s


AnsiCode.__call__ = apply_ansi  # type: ignore
AnsiCode.wrap = wrap_ansi  # type: ignore


####################################################################################################


# fmt: off
RESET    = AnsiCode.RESET    = AnsiCode('0')
RESET_BG = AnsiCode.RESET_BG = AnsiCode('49')
RESET_FG = AnsiCode.RESET_FG = AnsiCode('39')
WARN     = AnsiCode.WARN     = AnsiCode('1;43;30')
ERROR    = AnsiCode.ERROR    = AnsiCode('1;41;30')
WHITE    = AnsiCode.WHITE    = AnsiCode('37')
RED      = AnsiCode.RED      = AnsiCode('31')
GREEN    = AnsiCode.GREEN    = AnsiCode('32')
YELLOW   = AnsiCode.YELLOW   = AnsiCode('33')
BLUE     = AnsiCode.BLUE     = AnsiCode('34')
MAGENTA  = AnsiCode.MAGENTA  = AnsiCode('35')
CYAN     = AnsiCode.CYAN     = AnsiCode('36')
BOLD     = AnsiCode.BOLD     = AnsiCode('1')
DIM      = AnsiCode.DIM      = AnsiCode('2')
ITALIC   = AnsiCode.ITALIC   = AnsiCode('3')
UNDER    = AnsiCode.UNDER    = AnsiCode('4')
GRAY_BG  = AnsiCode.GRAY_BG  = AnsiCode('37;100')
ICON     = AnsiCode.ICON     = AnsiCode('1;97;100')
# fmt: on
